# Copyright 2004-2009 Joe Wreschnig, Michael Urman, Steven Robertson
#           2011-2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import random
import re
import ctypes
import ctypes.util
import sys
import unicodedata
import threading
import locale
from collections.abc import Callable
from functools import reduce

# Windows doesn't have fcntl, just don't lock for now
try:
    import fcntl

    fcntl
except ImportError:
    fcntl = None  # type: ignore

from senf import fsnative

from quodlibet.util.string.titlecase import title

from quodlibet.const import SUPPORT_EMAIL, COPYRIGHT
from quodlibet.util.dprint import print_d, print_, print_e, print_w, print_exc
from .misc import cached_func, get_module_dir, get_ca_file, NamedTemporaryFile, cmp
from .environment import (
    is_plasma,
    is_unity,
    is_enlightenment,
    is_linux,
    is_windows,
    is_wine,
    is_osx,
    is_flatpak,
    matches_flatpak_runtime,
)
from .enum import enum
from .i18n import _, ngettext, C_


# flake8
(
    cached_func,
    enum,
    print_w,
    print_exc,
    is_plasma,
    is_unity,
    is_enlightenment,
)
(
    is_linux,
    is_windows,
    is_wine,
    is_osx,
    get_module_dir,
    get_ca_file,
)
NamedTemporaryFile, is_flatpak, cmp, matches_flatpak_runtime


class InstanceTracker:
    """A mixin for GObjects to return a list of all alive objects
    of a given type. Note that it must be used with a GObject or
    something with a connect method and destroy signal."""

    __kinds: dict[type, list[object]] = {}

    def _register_instance(self, klass=None):
        """Register this object to be returned in the active instance list."""
        if klass is None:
            klass = type(self)
        self.__kinds.setdefault(klass, []).append(self)
        self.connect("destroy", self.__kinds[klass].remove)

    @classmethod
    def instances(cls):
        return cls.__kinds.get(cls, [])


class OptionParser:
    def __init__(self, name, version, description=None, usage=None):
        self.__name = name
        self.__version = version
        self.__args = {}
        self.__translate_short = {}
        self.__translate_long = {}
        self.__help = {}
        self.__usage = usage
        self.__description = description
        self.add("help", shorts="h", help=_("Display brief usage information"))
        self.add("version", shorts="v", help=_("Display version and copyright"))
        self.add("debug", shorts="d", help=_("Print debugging information"))

    def add(self, canon, help=None, arg="", shorts="", longs=None):
        longs = longs or []
        self.__args[canon] = arg
        for s in shorts:
            self.__translate_short[s] = canon
        for l in longs:
            self.__translate_long[l] = canon
        if help:
            self.__help[canon] = help

    def __shorts(self):
        shorts = ""
        for short, canon in self.__translate_short.items():
            shorts += short + (self.__args[canon] and "=" or "")
        return shorts

    def __longs(self):
        longs = []
        for long_, arg in self.__args.items():
            longs.append(long_ + (arg and "=" or ""))
        for long_, canon in self.__translate_long.items():
            longs.append(long_ + (self.__args[canon] and "=" or ""))
        return longs

    def __format_help(self, opt, space):
        if opt in self.__help:
            help = self.__help[opt]
            if self.__args[opt]:
                opt = f"{opt}={self.__args[opt]}"
            return f"  --{opt.ljust(space)} {help}\n"
        return ""

    def help(self):
        l = 0
        for k in self.__help.keys():
            l = max(l, len(k) + len(self.__args.get(k, "")) + 4)

        s = _("Usage: %(program)s %(usage)s") % {
            "program": sys.argv[0],
            "usage": self.__usage if self.__usage else _("[options]"),
        }
        s += "\n"
        if self.__description:
            s += f"{self.__name} - {self.__description}\n"
        s += "\n"
        keys = sorted(self.__help.keys())
        try:
            keys.remove("help")
        except ValueError:
            pass
        try:
            keys.remove("version")
        except ValueError:
            pass
        for h in keys:
            s += self.__format_help(h, l)
        if keys:
            s += "\n"
        s += self.__format_help("help", l)
        s += self.__format_help("version", l)
        return s

    def set_help(self, newhelp):
        self.__help = newhelp

    def version(self):
        return f"""\
{self.__name} {self.__version}
<{SUPPORT_EMAIL}>
{COPYRIGHT}\
"""

    def parse(self, args=None):
        if args is None:
            args = sys.argv[1:]
        from getopt import getopt, GetoptError

        try:
            opts, args = getopt(args, self.__shorts(), self.__longs())
        except GetoptError as e:
            s = str(e)
            text = []
            if "not recognized" in s:
                text.append(_("Option %r not recognized.") % s.split()[1])
            elif "requires argument" in s:
                text.append(_("Option %r requires an argument.") % s.split()[1])
            elif "unique prefix" in s:
                text.append(_("%r is not a unique prefix.") % s.split()[1])
            if "help" in self.__args:
                text.append(_("Try %s --help.") % sys.argv[0])

            print_e("\n".join(text))
            raise SystemExit(True) from e
        else:
            transopts = {}
            for o, a in opts:
                if o.startswith("--"):
                    o = self.__translate_long.get(o[2:], o[2:])
                elif o.startswith("-"):
                    o = self.__translate_short.get(o[1:], o[1:])
                if o == "help":
                    print_(self.help())
                    raise SystemExit
                if o == "version":
                    print_(self.version())
                    raise SystemExit
                if o == "debug":
                    from quodlibet import const

                    const.DEBUG = True
                if self.__args[o]:
                    transopts[o] = a
                else:
                    transopts[o] = True

            return transopts, args


def escape(string: str) -> str:
    """Escape a string in a manner suitable for XML/Pango."""
    return string.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")


def unescape(string: str) -> str:
    """Unescape a string in a manner suitable for XML/Pango."""
    return string.replace("&lt;", "<").replace("&gt;", ">").replace("&amp;", "&")


def bold(string: str, escaper: Callable[[str], str] = escape) -> str:
    return f"<b>{escaper(string)}</b>"


def monospace(string: str, escaper: Callable[[str], str] = escape) -> str:
    return f"<tt>{escaper(string)}</tt>"


def italic(string: str, escaper: Callable[[str], str] = escape) -> str:
    return f"<i>{escaper(string)}</i>"


def bold_italic(string: str, escaper: Callable[[str], str] = escape) -> str:
    return f"<b><i>{escaper(string)}</i></b>"


def parse_time(time_str: str, suppress_errors: bool = True):
    """Parse a time string in hh:mm:ss, mm:ss, or ss format."""

    if time_str[0:1] == "-":
        m = -1
        time_str = time_str[1:]
    else:
        m = 1

    try:
        return m * reduce(lambda s, a: s * 60 + int(a), re.split(r":|\.", time_str), 0)
    except (ValueError, re.error):
        if not suppress_errors:
            raise
        return 0


def validate_query_date(datestr):
    """Validates a user provided date that can be compared using date_key().

    Returns True id the date is valid.
    """

    parts = datestr.split("-")
    if len(parts) > 3:
        return False

    if len(parts) > 2:
        try:
            v = int(parts[2])
        except ValueError:
            return False
        else:
            if not 1 <= v <= 31:
                return False

    if len(parts) > 1:
        try:
            v = int(parts[1])
        except ValueError:
            return False
        else:
            if not 1 <= v <= 12:
                return False

    try:
        int(parts[0])
    except ValueError:
        return False

    return True


def date_key(datestr):
    """Parse a date format y-m-d and returns an undefined integer that
    can only be used to compare dates.

    In case the date string is invalid the returned value is undefined.
    """

    # this basically does "2001-02-03" -> 20010203

    default = [0, 1, 1]
    parts = datestr.split("-")
    parts += default[len(parts) :]

    value = 0
    for d, p, m in zip(default, parts, (10000, 100, 1), strict=False):
        try:
            value += int(p) * m
        except ValueError:
            # so that "2003-01-" is equal to "2003-01" ..
            value += d * m
    return value


def parse_date(datestr):
    """Parses yyyy-mm-dd date format and returns unix time.

    Raises ValueError in case the input couldn't be parsed.
    """

    import time

    try:
        frmt = ["%Y", "%Y-%m", "%Y-%m-%d"][datestr.count("-")]
    except IndexError as e:
        raise ValueError from e

    try:
        return time.mktime(time.strptime(datestr, frmt))
    except OverflowError as e:
        raise ValueError(e) from e


def parse_year(yearstr):
    """Parses dates of various formats and returns the year."""
    if "/" in yearstr:
        return yearstr[-4:]
    return yearstr[:4]


def format_int_locale(value):
    """Turn an integer into a grouped, locale-dependent string
    e.g. 12345 -> "12,345" or "12.345" etc
    """
    return locale.format_string("%d", value, grouping=True)


def format_float_locale(value, format="%.2f"):
    """Turn a float into a grouped, locale-dependent string
    e.g. 12345.67 -> "12,345.67" or "12.345,67" etc
    """
    return locale.format_string(format, value, grouping=True)


def format_rating(value, blank=True):
    """Turn a number into a sequence of rating symbols."""

    from quodlibet import config

    prefs = config.RATINGS
    steps = prefs.number
    value = max(min(value, 1.0), 0)
    ons = int(round(steps * value))
    offs = (steps - ons) if blank else 0
    return prefs.full_symbol * ons + prefs.blank_symbol * offs


def format_bitrate(value):
    return _("%d kbps") % int(value)


def format_size(size):
    """Turn an integer size value into something human-readable.

    Args:
        size (int): size in bytes
    Returns:
        str
    """
    # TODO: Better i18n of this (eg use O/KO/MO/GO in French)
    if size >= 1024**3:
        return "%.1f GB" % (float(size) / (1024**3))
    if size >= 1024**2 * 100:
        return "%.0f MB" % (float(size) / (1024**2))
    if size >= 1024**2 * 10:
        return "%.1f MB" % (float(size) / (1024**2))
    if size >= 1024**2:
        return "%.2f MB" % (float(size) / (1024**2))
    if size >= 1024 * 10:
        return "%d KB" % int(size / 1024)
    if size >= 1024:
        return "%.2f KB" % (float(size) / 1024)
    return "%d B" % size


def format_time(time):
    """Turn a time value in seconds into hh:mm:ss or mm:ss."""

    if time < 0:
        time = abs(time)
        prefix = "-"
    else:
        prefix = ""
    if time >= 3600:  # 1 hour
        # time, in hours:minutes:seconds
        return "%s%d:%02d:%02d" % (prefix, time // 3600, (time % 3600) // 60, time % 60)
    # time, in minutes:seconds
    return "%s%d:%02d" % (prefix, time // 60, time % 60)


def format_time_display(time):
    """Like format_time, but will use RATIO instead of a colon to separate"""

    return format_time(time).replace(":", "\u2236")


def format_time_seconds(time):
    from quodlibet import ngettext

    time_str = format_int_locale(time)
    return ngettext("%s second", "%s seconds", time) % time_str


def format_time_long(time, limit=2):
    """Turn a time value in seconds into x hours, x minutes, etc.

    `limit` limits the count of units used, so the result will be <= time.
    0 means no limit.
    """

    from quodlibet import ngettext

    if time < 1:
        return _("No time information")

    cutoffs = [
        (60, lambda n: ngettext("%d second", "%d seconds", n)),
        (60, lambda n: ngettext("%d minute", "%d minutes", n)),
        (24, lambda n: ngettext("%d hour", "%d hours", n)),
        (365, lambda n: ngettext("%d day", "%d days", n)),
        (None, lambda n: ngettext("%d year", "%d years", n)),
    ]

    time_str = []
    for divisor, gettext_partial in cutoffs:
        if time < 1:
            break
        if divisor is None:
            time, unit = 0, time
        else:
            time, unit = divmod(time, divisor)
        if unit:
            time_str.append(gettext_partial(unit) % unit)
    time_str.reverse()

    if limit:
        time_str = time_str[:limit]

    return ", ".join(time_str)


def format_time_preferred(t, fmt=None):
    """Returns duration formatted to user's preference"""
    from quodlibet.config import DurationFormat, DURATION

    fmt = fmt or DURATION.format

    if fmt == DurationFormat.STANDARD:
        return format_time_long(t, 2)
    if fmt == DurationFormat.NUMERIC:
        return format_time_display(t)
    if fmt == DurationFormat.FULL:
        return format_time_long(t, 5)
    return format_time_seconds(t)


def capitalize(str):
    """Capitalize a string, not affecting any character after the first."""
    return str[:1].upper() + str[1:]


def _split_numeric_sortkey(
    s, limit=10, reg=re.compile(r"[0-9][0-9]*\.?[0-9]*").search, join=" ".join
):
    """Separate numeric values from the string and convert to float, so
    it can be used for human sorting. Also removes all extra whitespace."""
    result = reg(s)
    if not result or not limit:
        text = join(s.split())
        return (text,) if text else ()
    start, end = result.span()
    return (
        join(s[:start].split()),
        float(result.group()),
        _split_numeric_sortkey(s[end:], limit - 1),
    )


def human_sort_key(s, normalize=unicodedata.normalize):
    if not s:
        return ()
    if not isinstance(s, str):
        s = s.decode("utf-8")
    s = normalize("NFD", s.lower())
    return _split_numeric_sortkey(s)


def website(site):
    """Open the given URL in the user's default browser"""

    from gi.repository import Gtk, Gdk, GLib

    try:
        Gtk.show_uri(None, site, Gdk.CURRENT_TIME)
    except GLib.Error:
        print_exc()


def tag(name: str, cap: bool = True) -> str:
    # Return a 'natural' version of the tag for human-readable bits.
    # Strips ~ and ~# from the start and runs it through a map (which
    # the user can configure).
    if not name:
        return ngettext("Invalid tag", "Invalid tags", 1)
    from quodlibet.util.tags import readable

    parts = map(readable, tagsplit(name))
    if cap:
        # Translators: If tag names, when capitalized, should not
        # be title-cased ("Looks Like This"), but rather only have
        # the first letter capitalized, translate this string as
        # something non-empty other than "titlecase?".
        if C_("check", "titlecase?") == "titlecase?":
            parts = map(title, parts)
        else:
            parts = map(capitalize, parts)
    return " / ".join(parts)


def tagsplit(tag: str) -> list[str]:
    """Split a (potentially) tied tag into a list of atomic tags. Two ~~s
    make the next tag prefixed with a ~, so ~foo~~bar => [foo, ~bar]."""
    if "~" in tag[1:]:
        if tag.startswith("~") and not tag.startswith("~#"):
            tag = tag[1:]
        tags = []
        front = ""
        for part in tag.split("~"):
            if part:
                tags.append(front + part)
                front = ""
            else:
                front = "~"
        return tags
    return [tag]


def pattern(pat, cap=True, esc=False, markup=False):
    """Return a 'natural' version of the pattern string for human-readable
    bits. Assumes all tags in the pattern are present.
    """

    from quodlibet.pattern import Pattern, XMLFromPattern, XMLFromMarkupPattern
    from quodlibet.formats import FILESYSTEM_TAGS

    class Fakesong(dict):
        cap = False

        def comma(self, key):
            return " - ".join(self.list(key))

        def list(self, key):
            return [tag(k, self.cap) for k in tagsplit(key)]

        list_separate = list

        def __call__(self, tag, *args):
            if tag in FILESYSTEM_TAGS:
                return fsnative(str(tag))
            return 0 if "~#" in tag[:2] else self.comma(tag)

    fakesong = Fakesong({"filename": tag("filename", cap)})
    fakesong.cap = cap
    try:
        if markup:
            p = XMLFromMarkupPattern(pat)
        elif esc:
            p = XMLFromPattern(pat)
        else:
            p = Pattern(pat)
    except ValueError:
        return _("Invalid pattern")

    return p.format(fakesong)


def spawn(argv, stdout=False):
    """Asynchronously run a program. argv[0] is the executable name, which
    must be fully qualified or in the path. If stdout is True, return
    a file object corresponding to the child's standard output; otherwise,
    return the child's process ID.

    argv must be strictly str objects to avoid encoding confusion.
    """

    from gi.repository import GLib

    print_d(f"Running {argv!r}")
    args = GLib.spawn_async(
        argv=argv, flags=GLib.SpawnFlags.SEARCH_PATH, standard_output=stdout
    )

    if stdout:
        return os.fdopen(args[2])
    return args[0]


def fver(tup):
    return ".".join(map(str, tup))


def make_case_insensitive(filename):
    return "".join([f"[{c.lower()}{c.upper()}]" for c in filename])


class DeferredSignal:
    """A wrapper for connecting functions to signals.

    Some signals may fire hundreds of times, but only require processing
    once per group. This class pushes the call to the mainloop at idle
    priority and prevents multiple calls from being inserted in the
    mainloop at a time, greatly improving responsiveness in some places.

    When the target function is finally called, the arguments passed
    are the last arguments passed to DeferredSignal.

    `priority` defaults to GLib.PRIORITY_DEFAULT

    If `owner` is given, it will not call the target after the owner is
    destroyed.

    Example usage:

    def func(widget, user_arg):
        pass
    widget.connect('signal', DeferredSignal(func, owner=widget), user_arg)
    """

    def __init__(self, func, timeout=None, owner=None, priority=None):
        """timeout in milliseconds"""

        self.func = func
        self.dirty = False
        self.args = None

        if owner:

            def destroy_cb(owner):
                self.abort()

            owner.connect("destroy", destroy_cb)

        from gi.repository import GLib

        if priority is None:
            priority = GLib.PRIORITY_DEFAULT

        if timeout is None:
            self.do_idle_add = lambda f: GLib.idle_add(f, priority=priority)
        else:
            self.do_idle_add = lambda f: GLib.timeout_add(timeout, f, priority=priority)

    @property
    def __self__(self):
        return self.func.__self__

    @property
    def __code__(self):
        return self.func.__code__

    @property
    def __closure__(self):
        return self.func.__closure__

    def call(self, *args):
        """Force a call"""

        self.abort()
        self.args = args
        self._wrap()

    def abort(self):
        """Abort any queued up calls.

        Can still be reused afterwards.
        """

        if self.dirty:
            from gi.repository import GLib

            GLib.source_remove(self._id)
            self.dirty = False
            self.args = None

    def __call__(self, *args):
        self.args = args
        if not self.dirty:
            self.dirty = True
            self._id = self.do_idle_add(self._wrap)

    def _wrap(self):
        self.func(*self.args)
        self.dirty = False
        self.args = None
        return False


def connect_obj(this, detailed_signal, handler, that, *args, **kwargs):
    """A wrapper for connect() that has the same interface as connect_object().
    Used as a temp solution to get rid of connect_object() calls which may
    be changed to match the C version more closely in the future.

    https://git.gnome.org/browse/pygobject/commit/?id=86fb12b3e9b75

    While it's not clear if switching to weak references will break anything,
    we mainly used this for adjusting the callback signature. So using
    connect() behind the scenes will keep things working as they are now.
    """

    def wrap(this, *args):
        return handler(that, *args)

    return this.connect(detailed_signal, wrap, *args, **kwargs)


def _connect_destroy(sender, func, detailed_signal, handler, *args, **kwargs):
    """Connect a bound method to a foreign object signal and disconnect
    if the object the method is bound to emits destroy (Gtk.Widget subclass).

    Also works if the handler is a nested function in a method and
    references the method's bound object.

    This solves the problem that the sender holds a strong reference
    to the bound method and the bound to object doesn't get GCed.
    """

    if hasattr(handler, "__self__"):
        obj = handler.__self__
    else:
        # XXX: get the "self" var of the enclosing scope.
        # Used for nested functions which ref the object but aren't methods.
        # In case they don't ref "self" normal connect() should be used anyway.
        index = handler.__code__.co_freevars.index("self")
        obj = handler.__closure__[index].cell_contents

    assert obj is not sender

    handler_id = func(detailed_signal, handler, *args, **kwargs)

    def disconnect_cb(*args):
        sender.disconnect(handler_id)

    obj.connect("destroy", disconnect_cb)
    return handler_id


def connect_destroy(sender, *args, **kwargs):
    return _connect_destroy(sender, sender.connect, *args, **kwargs)


def connect_after_destroy(sender, *args, **kwargs):
    return _connect_destroy(sender, sender.connect_after, *args, **kwargs)


class cached_property:  # noqa
    """A read-only @property that is only evaluated once.
    TODO: work out some typing for this, see test_query.py"""

    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = name = fget.__name__
        # these get name mangled, so caching won't work unless
        # we mangle too
        assert not (
            name.startswith("__") and not name.endswith("__")
        ), "can't cache a dunder method"

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


def sanitize_tags(tags, stream=False):
    """Returns a new sanitized tag dict. stream defines if the
    tags of a main/base song should be changed or of a stream song.
    e.g. title will be removed for the base song but not for the stream one.
    """

    san = {}
    for key, value in tags.items():
        key = key.lower()
        key = {"location": "website"}.get(key, key)

        if isinstance(value, str):
            lower = value.lower().strip()

            if key == "channel-mode":
                if "stereo" in lower or "dual" in lower:
                    value = "stereo"
            elif key == "audio-codec":
                if "mp3" in lower:
                    value = "MP3"
                elif "aac" in lower or "advanced" in lower:
                    value = "MPEG-4 AAC"
                elif "vorbis" in lower:
                    value = "Ogg Vorbis"

            if lower in (
                "http://www.shoutcast.com",
                "http://localhost/",
                "default genre",
                "none",
                "http://",
                "unnamed server",
                "unspecified",
                "n/a",
            ):
                continue

        if key == "duration":
            try:
                value = int(int(value) / 1000)
            except ValueError:
                pass
            else:
                if not stream:
                    continue
                key = "~#length"
        elif key == "bitrate":
            try:
                value = int(value) / 1000
            except ValueError:
                pass
            else:
                if not stream:
                    continue
                key = "~#bitrate"
        elif key == "nominal-bitrate":
            try:
                value = int(value) / 1000
            except ValueError:
                pass
            else:
                if stream:
                    continue
                key = "~#bitrate"

        if key in (
            "emphasis",
            "mode",
            "layer",
            "maximum-bitrate",
            "minimum-bitrate",
            "has-crc",
            "homepage",
        ):
            continue

        if not stream and key in ("title", "album", "artist", "date"):
            continue

        if isinstance(value, int | float):
            if not key.startswith("~#"):
                key = "~#" + key
            san[key] = value
        else:
            if key.startswith("~#"):
                key = key[2:]

            if not isinstance(value, str):
                continue

            value = value.strip()
            if key in san:
                if value not in san[key].split("\n"):
                    san[key] += "\n" + value
            else:
                san[key] = value

    return san


def build_filter_query(key, values):
    """Create a text query that matches a union of all values for a key

    build_filter_query("foo", ["x", "y"])
    => foo = |("x"c, "y"c)
    build_filter_query("~#foo", ["1"])
    => #(foo = 1)
    """

    if not values:
        return ""
    if key.startswith("~#"):
        nheader = key[2:]
        queries = [f"#({nheader} = {i})" for i in values]
        if len(queries) > 1:
            return "|({})".format(", ".join(queries))
        return queries[0]
    text = ", ".join(
        ["'{}'c".format(v.replace("\\", "\\\\").replace("'", "\\'")) for v in values]
    )
    if len(values) == 1:
        return f"{key} = {text}"
    return f"{key} = |({text})"


def limit_songs(songs, max, weight_by_ratings=False):
    """Choose at most `max` songs from `songs`,
    optionally giving weighting to ~#rating"""

    if not max or len(songs) < max:
        return songs
    if weight_by_ratings:

        def rating_weighted_random(song):
            # Apply even (random : higher rating) weighting
            return (1 - song("~#rating")) * (1 + random.random())

        songs.sort(key=rating_weighted_random)
    else:
        random.shuffle(songs)
    return songs[:max]


def gi_require_versions(name, versions):
    """Like gi.require_version, but will take a list of versions.

    Returns the required version or raises ValueError.
    """

    assert versions

    import gi

    error = None
    for version in versions:
        try:
            gi.require_version(name, version)
        except ValueError as e:
            error = e
        else:
            return version
    else:
        raise error


def load_library(names, shared=True):
    """Load a ctypes library with a range of names to try.

    Handles direct .so names and library names ["libgpod.so", "gpod"].

    If shared is True can return a shared instance.
    Raises OSError if not found.

    Returns (library, name)
    """

    if not names:
        raise ValueError

    if shared:

        def load_func(n):
            return getattr(ctypes.cdll, n)
    else:
        load_func = ctypes.cdll.LoadLibrary

    errors = []
    for name in names:
        dlopen_name = name
        if ".so" not in name and ".dll" not in name and ".dylib" not in name:
            dlopen_name = ctypes.util.find_library(name) or name

        if is_osx() and not os.path.isabs(dlopen_name):
            dlopen_name = os.path.join(sys.prefix, "lib", dlopen_name)

        try:
            return load_func(dlopen_name), name
        except (OSError, AttributeError) as e:
            errors.append(str(e))

    raise OSError("\n".join(errors))


def is_main_thread():
    """If the calling thread is the main one"""

    return threading.current_thread().name == "MainThread"


class MainRunnerError(Exception):
    pass


class MainRunnerAbortedError(MainRunnerError):
    pass


class MainRunnerTimeoutError(MainRunnerError):
    pass


class MainRunner:
    """Schedule a function call in the main loop from a
    worker thread and wait for the result.

    Make sure to call abort() before the main loop gets destroyed, otherwise
    the worker thread may block forever in call().
    """

    def __init__(self):
        self._source_id = None
        self._call_id = None
        self._lock = threading.Lock()
        self._cond = threading.Condition(self._lock)
        self._return = None
        self._error = None
        self._aborted = False

    def _run(self, func, *args, **kwargs):
        try:
            self._return = func(*args, **kwargs)
        except Exception as e:
            self._error = MainRunnerError(e)

    def _idle_run(self, call_id, call_event, func, *args, **kwargs):
        call_event.set()
        with self._lock:
            # In case a timeout happened but this got still
            # scheduled, this could be called after call() returns;
            # Compare to the current call id and do nothing if it isn't ours
            if call_id is not self._call_id:
                return False
            try:
                self._run(func, *args, **kwargs)
            finally:
                self._source_id = None
                self._cond.notify()
            return False

    def abort(self):
        """After this call returns no function will be executed anymore
        and a currently blocking call will fail with MainRunnerAbortedError.

        Can be called multiple times and can not fail.
        call() will always fail after this was called.
        """

        from gi.repository import GLib

        with self._lock:
            if self._aborted:
                return
            if self._source_id is not None:
                GLib.source_remove(self._source_id)
                self._source_id = None
            self._aborted = True
            self._call_id = None
            self._error = MainRunnerAbortedError("aborted")
            self._cond.notify()

    def call(self, func, *args, **kwargs):
        """Runs the function in the main loop and blocks until
        it is finished or abort() was called. In case this is called
        from the main loop the function gets executed immediately.

        The priority kwargs defines the event source priority and will
        not be passed to func.

        In case a timeout kwarg is given the call will raise
        MainRunnerTimeoutError in case the function hasn't been scheduled
        (doesn't mean returned) until that time. timeout is a float in seconds.

        Can raise MainRunnerError in case the function raises an exception.
        Raises MainRunnerAbortedError in case the runner was aborted.
        Raises MainRunnerTimeoutError in case the timeout was reached.
        """

        from gi.repository import GLib

        with self._lock:
            if self._aborted:
                raise self._error
            self._error = None
            # XXX: ideally this should be GLib.MainContext.default().is_owner()
            # but that's not available in older pygobject
            if is_main_thread():
                kwargs.pop("priority", None)
                self._run(func, *args, **kwargs)
            else:
                assert self._source_id is None
                assert self._call_id is None
                timeout = kwargs.pop("timeout", None)
                call_event = threading.Event()
                self._call_id = object()
                self._source_id = GLib.idle_add(
                    self._idle_run, self._call_id, call_event, func, *args, **kwargs
                )
                # only wait for the result if we are sure it got scheduled
                if call_event.wait(timeout):
                    self._cond.wait()
                self._call_id = None
                if self._source_id is not None:
                    GLib.source_remove(self._source_id)
                    self._source_id = None
                    raise MainRunnerTimeoutError(f"timeout: {timeout!r}")
            if self._error is not None:
                raise self._error
            return self._return


def re_escape(string, bad="/.^$*+-?{,\\[]|()<>#=!:"):
    """A `re.escape` which also works with unicode"""

    def needs_escape(c):
        return c in bad and "\\" + c or c

    return type(string)().join(map(needs_escape, string))


PR_SET_NAME = 15


def set_process_title(title):
    """Sets process name as visible in ps or top. Requires ctypes libc
    and is almost certainly *nix-only. See issue 736
    """

    if os.name == "nt":
        return

    try:
        libc = load_library(["libc.so.6", "c"])[0]
        prctl = libc.prctl
    except (OSError, AttributeError):
        print_d("Couldn't find module libc.so.6 (ctypes). Not setting process title.")
    else:
        prctl.argtypes = [
            ctypes.c_int,
            ctypes.c_ulong,
            ctypes.c_ulong,
            ctypes.c_ulong,
            ctypes.c_ulong,
        ]
        prctl.restype = ctypes.c_int
        data = ctypes.create_string_buffer(title.encode("utf-8"))
        res = prctl(PR_SET_NAME, ctypes.addressof(data), 0, 0, 0)
        if res != 0:
            print_w("Setting the process title failed")


def list_unique(sequence):
    """Takes any sequence and returns a list with all duplicate entries
    removed while preserving the order.
    """

    l = []
    seen = set()
    append = l.append
    add = seen.add
    for v in sequence:
        if v not in seen:
            append(v)
            add(v)
    return l


def reraise(tp, value, tb=None):
    """Reraise an exception with a new exception type and
    the original stack trace
    """

    if tb is None:
        tb = sys.exc_info()[2]
    raise tp(value).with_traceback(tb)



# Copyright 2018 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import print_d
from ._base import SessionClient, SessionError


class XSMPSessionClient(SessionClient):
    def __init__(self):
        super().__init__()
        self._client = None

    def open(self, app):
        try:
            from ._xsmp import XSMPClient, XSMPError
        except ImportError as e:
            raise SessionError(e) from e

        print_d("Connecting with XSMP")
        client = XSMPClient()
        try:
            client.open()
        except XSMPError as e:
            raise SessionError(e) from e

        try:
            from gi.repository import GdkX11
        except ImportError:
            pass
        else:
            GdkX11.x11_set_sm_client_id(client.client_id)

        print_d(f"Connected. Client ID: {client.client_id}")

        def save_yourself(client, *args):
            print_d(f"xsmp: save_yourself {args!r}")
            client.save_yourself_done(True)

        def die(client, *args):
            print_d("xsmp: die")
            app.quit()

        def save_complete(client):
            print_d("xsmp: save_complete")

        def shutdown_cancelled(client):
            print_d("xsmp: shutdown_cancelled")

        client.connect("save-yourself", save_yourself)
        client.connect("die", die)
        client.connect("save-complete", save_complete)
        client.connect("shutdown-cancelled", shutdown_cancelled)
        self._client = client

    def close(self):
        if self._client is None:
            return

        print_d("Disconnecting from XSMP")
        self._client.close()
        self._client = None



# Copyright 2014 Christoph Reiter
#           2016 Nick Boultbee
#           2020 Daniel Petrescu
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""A list of icon names.

Mainly to see which ones we define and for looking up the standardized ones.
Also reduces the chance of making a typo.

http://standards.freedesktop.org/icon-naming-spec/icon-naming-spec-latest.html
https://docs.google.com/spreadsheet/pub?
    key=0AsPAM3pPwxagdGF4THNMMUpjUW5xMXZfdUNzMXhEa2c
"""

from quodlibet.util.enum import enum


@enum
class Icons(str):
    NONE = ""

    # these we have in our fallback icon theme
    QUODLIBET = "io.github.quodlibet.QuodLibet"
    EXFALSO = "exfalso"
    QUODLIBET_MISSING_COVER = "quodlibet-missing-cover"

    # stock icon I couldn't think of a replacement for now
    EDIT = "gtk-edit"

    # these only come from an icon theme
    APPLICATION_EXIT = "application-exit"  # "_Quit"
    APPLICATION_INTERNET = "applications-internet"
    APPLICATION_UTILITIES = "applications-utilities"
    APPOINTMENT_NEW = "appointment-new"
    AUDIO_CARD = "audio-card"
    AUDIO_INPUT_MICROPHONE = "audio-input-microphone"
    AUDIO_VOLUME_MUTED = "audio-volume-muted"
    AUDIO_X_GENERIC = "audio-x-generic"
    CHANGES_PREVENT = "changes-prevent"
    DIALOG_ERROR = "dialog-error"  # "Error"
    DIALOG_INFORMATION = "dialog-information"  # "Information"
    DIALOG_PASSWORD = "dialog-password"
    DIALOG_QUESTION = "dialog-question"  # "Question"
    DIALOG_WARNING = "dialog-warning"  # "Warning"
    DOCUMENT_NEW = "document-new"  # "_New"
    DOCUMENT_OPEN = "document-open"  # "_Open"
    DOCUMENT_PROPERTIES = "document-properties"  # "_Properties"
    DOCUMENT_REVERT = "document-revert"  # "_Revert"
    DOCUMENT_SAVE = "document-save"  # "_Save"
    DOCUMENT_SAVE_AS = "document-save-as"  # "_Save"
    DOCUMENT_SEND = "document-send"
    DRIVE_REMOVABLE_MEDIA = "drive-removable-media"
    EDIT_CLEAR = "edit-clear"  # "_Clear"
    EDIT_COPY = "edit-copy"  # "_Copy"
    EDIT_PASTE = "edit-paste"  # "_Paste"
    EDIT_DELETE = "edit-delete"  # "_Delete"
    EDIT_FIND = "edit-find"  # "_Find"
    EDIT_FIND_REPLACE = "edit-find-replace"  # "Find and _Replace"
    EDIT_REDO = "edit-redo"  # "_Redo"
    EDIT_SELECT_ALL = "edit-select-all"
    EDIT_UNDO = "edit-undo"  # "_Undo"
    EMBLEM_DOWNLOADS = "emblem-downloads"
    EMBLEM_SHARED = "emblem-shared"
    EMBLEM_SYSTEM = "emblem-system"
    FACE_SMILE = "face-smile"
    FAVORITE = "emblem-favorite"
    FOLDER = "folder"
    FOLDER_DRAG_ACCEPT = "folder-drag-accept"
    FOLDER_OPEN = "folder-open"
    FORMAT_JUSTIFY_FILL = "format-justify-fill"
    GO_JUMP = "go-jump"  # "_Jump to"
    HELP_ABOUT = "help-about"  # "_About"
    HELP_BROWSER = "help-browser"  # "_Help"
    IMAGE_X_GENERIC = "image-x-generic"
    INSERT_IMAGE = "insert-image"
    INSERT_TEXT = "insert-text"
    LIST_ADD = "list-add"  # "_Add"
    LIST_EDIT = "document-edit"  # "_Edit"
    LIST_REMOVE = "list-remove"  # "_Remove"
    MEDIA_EJECT = "media-eject"
    MEDIA_OPTICAL = "media-optical"  # "_CD-ROM"
    MEDIA_PLAYBACK_PAUSE = "media-playback-pause"  # "P_ause"
    MEDIA_PLAYBACK_START = "media-playback-start"  # "_Play"
    MEDIA_PLAYBACK_STOP = "media-playback-stop"  # "_Stop"
    MEDIA_PLAYLIST_REPEAT = "media-playlist-repeat"
    MEDIA_PLAYLIST_SHUFFLE = "media-playlist-shuffle"
    MEDIA_RECORD = "media-record"
    MEDIA_SKIP_BACKWARD = "media-skip-backward"  # "Pre_vious"
    MEDIA_SKIP_FORWARD = "media-skip-forward"  # "_Next"
    MULTIMEDIA_PLAYER = "multimedia-player"
    MULTIMEDIA_VOLUME_CONTROL = "multimedia-volume-control"
    NETWORK_WORKGROUP = "network-workgroup"  # "_Network"
    NETWORK_SERVER = "network-server"
    NETWORK_TRANSMIT = "network-transmit"
    PREFERENCES_SYSTEM = "preferences-system"  # "_Preferences"
    PREFERENCES_DESKTOP_SCREENSAVER = "preferences-desktop-screensaver"
    PREFERENCES_DESKTOP_THEME = "preferences-desktop-theme"
    PROCESS_STOP = "process-stop"  # "_Stop"
    SYSTEM_LOCK_SCREEN = "system-lock-screen"
    SYSTEM_RUN = "system-run"  # "_Execute"
    SYSTEM_SEARCH = "system-search"
    TEXT_HTML = "text-html"
    TOOLS_CHECK_SPELLING = "tools-check-spelling"  # "_Spell Check"
    USER_BOOKMARKS = "user-bookmarks"  # Looks like a rating
    USER_DESKTOP = "user-desktop"
    USER_TRASH = "user-trash"  # "Trash"
    UTILITIES_TERMINAL = "utilities-terminal"
    VIEW_LIST = "view-list"  #
    VIEW_REFRESH = "view-refresh"  # "_Refresh"
    WINDOW_CLOSE = "window-close"  # "_Close"



# Copyright 2017 Christoph Reiter
#           2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import platform

from gi.repository import Gtk, Pango
from senf import fsn2uri, fsn2text
import mutagen

from quodlibet import _
from quodlibet.qltk import Icons
from quodlibet.util.path import unexpand
from quodlibet.plugins.events import EventPlugin
from quodlibet import formats
from quodlibet import app, get_user_dir, get_cache_dir
from quodlibet.util import fver, escape
from quodlibet.qltk import (
    gtk_version,
    pygobject_version,
    get_backend_name,
    get_font_backend_name,
)
from quodlibet.qltk import show_uri


class AppInformation(EventPlugin):
    PLUGIN_ID = "AppInformation"
    PLUGIN_NAME = _("Application Information")
    PLUGIN_DESC = _("Various information about the application and its environment.")
    PLUGIN_CAN_ENABLE = False
    PLUGIN_ICON = Icons.PREFERENCES_SYSTEM

    def PluginPreferences(self, *args):
        vb = Gtk.VBox()

        row = 0
        grid = Gtk.Grid(column_spacing=12, row_spacing=6)

        def label_title(text):
            l = Gtk.Label(
                label=text,
                xalign=1,
                yalign=0,
                wrap=True,
                justify=Gtk.Justification.RIGHT,
                selectable=True,
            )
            l.get_style_context().add_class(Gtk.STYLE_CLASS_DIM_LABEL)
            return l

        def label_value(text):
            return Gtk.Label(
                label=text,
                wrap=True,
                xalign=0,
                yalign=0,
                width_chars=25,
                selectable=True,
            )

        def label_path(path):
            text = escape(fsn2text(unexpand(path)))
            l = Gtk.Label(
                label=f"<a href='{fsn2uri(path)}'>{text}</a>",
                use_markup=True,
                ellipsize=Pango.EllipsizeMode.MIDDLE,
                xalign=0,
                selectable=True,
            )

            l.connect("activate-link", show_uri)
            return l

        grid.insert_row(row)
        l = label_title(_("Supported Formats"))
        format_names = sorted([t.format for t in formats.types])
        v = label_value(", ".join(format_names))
        grid.attach(l, 0, row, 1, 1)
        grid.attach(v, 1, row, 1, 1)
        row += 1

        grid.insert_row(row)
        l = label_title(_("Configuration Directory"))
        v = label_path(get_user_dir())
        grid.attach(l, 0, row, 1, 1)
        grid.attach(v, 1, row, 1, 1)
        row += 1

        grid.insert_row(row)
        l = label_title(_("Cache Directory"))
        v = label_path(get_cache_dir())
        grid.attach(l, 0, row, 1, 1)
        grid.attach(v, 1, row, 1, 1)
        row += 1

        grid.insert_row(row)
        l = label_title(_("Audio Backend"))
        v = label_value(f"{app.player.name}\n{app.player.version_info}")
        grid.attach(l, 0, row, 1, 1)
        grid.attach(v, 1, row, 1, 1)
        row += 1

        grid.insert_row(row)
        l = label_title("Python")
        v = label_value(platform.python_version())
        grid.attach(l, 0, row, 1, 1)
        grid.attach(v, 1, row, 1, 1)
        row += 1

        grid.insert_row(row)
        l = label_title("Mutagen")
        v = label_value(fver(mutagen.version))
        grid.attach(l, 0, row, 1, 1)
        grid.attach(v, 1, row, 1, 1)
        row += 1

        grid.insert_row(row)
        l = label_title("Gtk+")
        v = label_value(
            f"{fver(gtk_version)} ({get_backend_name()}, {get_font_backend_name()})"
        )
        grid.attach(l, 0, row, 1, 1)
        grid.attach(v, 1, row, 1, 1)
        row += 1

        grid.insert_row(row)
        l = label_title("PyGObject")
        v = label_value(fver(pygobject_version))
        grid.attach(l, 0, row, 1, 1)
        grid.attach(v, 1, row, 1, 1)
        row += 1

        vb.pack_start(grid, True, True, 0)
        vb.show_all()

        return vb



# Copyright 2013 Christoph Reiter
#           2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import operator

from gi.repository import Gtk, Pango, Gdk

from quodlibet import qltk
from quodlibet.qltk.views import AllTreeView, TreeViewColumnButton
from quodlibet.qltk.songsmenu import SongsMenu
from quodlibet.qltk.properties import SongProperties
from quodlibet.qltk.information import Information
from quodlibet.qltk import is_accel
from quodlibet.util import connect_obj

from .models import PaneModel
from .util import PaneConfig


class Pane(AllTreeView):
    """Pane of the paned browser"""

    TARGET_INFO_QL = 1
    TARGET_INFO_URI_LIST = 2

    def __init__(self, library, prefs, next_=None):
        super().__init__()
        self.set_fixed_height_mode(True)

        self.config = PaneConfig(prefs)
        self.__next = next_
        self.__restore_values = None

        self.__no_fill = 0

        column = TreeViewColumnButton(title=self.config.title)

        def on_column_header_clicked(column, event):
            # In case the column header gets clicked select the "All" entry
            if (
                event.button != Gdk.BUTTON_PRIMARY
                or event.type != Gdk.EventType.BUTTON_PRESS
            ):
                return Gdk.EVENT_PROPAGATE
            self.set_selected([])
            return Gdk.EVENT_STOP

        column.set_clickable(True)
        column.connect("button-press-event", on_column_header_clicked)
        column.set_use_markup(True)
        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        column.set_fixed_width(60)

        render = Gtk.CellRendererText()
        render.set_property("ellipsize", Pango.EllipsizeMode.END)
        column.pack_start(render, True)

        def text_cdf(column, cell, model, iter_, data):
            entry = model.get_value(iter_)
            markup = entry.get_markup(self.config)
            cell.set_property("markup", markup)

        column.set_cell_data_func(render, text_cdf)

        render_count = Gtk.CellRendererText()
        render_count.set_property("xalign", 1.0)
        render_count.set_property("max-width-chars", 5)
        column.pack_end(render_count, True)
        # Tiny columns break too much rendering
        column.set_min_width(150)

        def count_cdf(column, cell, model, iter_, data):
            entry = model.get_value(iter_)
            markup = entry.get_count_markup(self.config)
            cell.markup = markup
            cell.set_property("markup", markup)

        column.set_cell_data_func(render_count, count_cdf)
        self.append_column(column)

        model = PaneModel(self.config)
        self.set_model(model)

        self.set_search_equal_func(self.__search_func, None)
        self.set_search_column(0)
        self.set_enable_search(True)

        selection = self.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        self.__sig = self.connect("selection-changed", self.__selection_changed)
        s = self.connect("popup-menu", self.__popup_menu, library)
        connect_obj(self, "destroy", self.disconnect, s)

        targets = [
            ("text/x-quodlibet-songs", Gtk.TargetFlags.SAME_APP, self.TARGET_INFO_QL),
            ("text/uri-list", 0, self.TARGET_INFO_URI_LIST),
        ]
        targets = [Gtk.TargetEntry.new(*t) for t in targets]

        self.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK, targets, Gdk.DragAction.COPY
        )
        self.connect("drag-data-get", self.__drag_data_get)
        self.connect("destroy", self.__destroy)

        librarian = library.librarian or library
        self.connect("key-press-event", self.__key_pressed, librarian)

    def __key_pressed(self, view, event, librarian):
        # if ctrl+a is pressed, intercept and select the All entry instead
        if is_accel(event, "<Primary>a"):
            self.set_selected([])
            return True
        if is_accel(event, "<Primary>Return", "<Primary>KP_Enter"):
            qltk.enqueue(self.__get_selected_songs(sort=True))
            return True
        if is_accel(event, "<alt>Return"):
            songs = self.__get_selected_songs(sort=True)
            if songs:
                window = SongProperties(librarian, songs, parent=self)
                window.show()
            return True
        if is_accel(event, "<Primary>I"):
            songs = self.__get_selected_songs(sort=True)
            if songs:
                window = Information(librarian, songs, self)
                window.show()
            return True
        return False

    def __repr__(self):
        return f"<{type(self).__name__} config={self.config!r}>"

    def parse_restore_string(self, config_value):
        assert isinstance(config_value, str)

        values = config_value.split("\t")[:-1]

        try:
            if int(values[0]):
                values[0] = None
            else:
                del values[0]
        except (ValueError, IndexError):
            pass

        self.__restore_values = values

    def get_restore_string(self):
        values = self.get_selected()

        # The first value tells us if All was selected
        all_ = None in values
        if all_:
            values.remove(None)
        all_ = str(int(bool(all_)))
        values = list(values)
        values.insert(0, all_)

        # The config lib strips all whitespace,
        # so add a bogus . at the end
        values.append(".")

        return "\t".join(values)

    @property
    def tags(self):
        """Tags this pane displays"""

        return self.config.tags

    def __destroy(self, *args):
        # needed for gc
        self.__next = None

    def __search_func(self, model, column, key, iter_, data):
        entry = model.get_value(iter_)
        return not entry.contains_text(key)

    def __drag_data_get(self, view, ctx, sel, tid, etime):
        songs = self.__get_selected_songs(sort=True)

        if tid == self.TARGET_INFO_QL:
            qltk.selection_set_songs(sel, songs)
        else:
            sel.set_uris([song("~uri") for song in songs])

    def __popup_menu(self, view, library):
        songs = self.__get_selected_songs(sort=True)
        menu = SongsMenu(library, songs)
        menu.show_all()
        return view.popup_menu(menu, 0, Gtk.get_current_event_time())

    def __selection_changed(self, *args):
        if self.__next:
            self.__next.fill(self.__get_selected_songs())

    def add(self, songs):
        self.get_model().add_songs(songs)

    def remove(self, songs, remove_if_empty=True):
        self.inhibit()
        self.get_model().remove_songs(songs, remove_if_empty)
        self.uninhibit()

    def matches(self, song):
        model, paths = self.get_selection().get_selected_rows()

        if not paths:
            return True

        return model.matches(paths, song)

    def inhibit(self):
        """Inhibit selection change events and song propagation"""

        self.__no_fill += 1
        self.handler_block(self.__sig)

    def uninhibit(self):
        """Uninhibit selection change events and song propagation"""

        self.handler_unblock(self.__sig)
        self.__no_fill -= 1

    def fill(self, songs):
        # Restore the selection
        if self.__restore_values is not None:
            selected = self.__restore_values
            self.__restore_values = None
        else:
            selected = self.get_selected()

        model = self.get_model()
        # If previously all entries were selected or None: select All
        if not selected or len(model) == len(selected):
            selected = [None]

        self.inhibit()
        with self.without_model():
            model.clear()
            model.add_songs(songs)

        self.set_selected(selected, jump=True)
        self.uninhibit()

        if self.__next and self.__no_fill == 0:
            self.__next.fill(self.__get_selected_songs())

    def scroll(self, song):
        """Select and scroll to entry which contains song"""

        def select_func(row):
            entry = row[0]
            return entry.contains_song(song)

        self.select_by_func(select_func, one=True)

    def list(self, tag):
        return self.get_model().list(tag)

    def get_selected(self):
        """A list of keys for selected entries"""

        model, paths = self.get_selection().get_selected_rows()
        return model.get_keys(paths)

    def set_selected(self, values, jump=False, force_any=True):
        """Select entries with key in values

        jump -- scroll the the first selected entry
        any -- if nothing gets selected, select the first entry
        """

        if self.get_model().is_empty():
            return

        values = values or []

        # If the selection is the same, change nothing
        if values != self.get_selected():
            self.inhibit()
            self.get_selection().unselect_all()

            def select_func(row):
                entry = row[0]
                return entry.key in values

            self.select_by_func(select_func, scroll=jump)
            self.uninhibit()

            self.get_selection().emit("changed")

        if force_any and self.get_selection().count_selected_rows() == 0:
            self.set_cursor((0,))

    def set_selected_by_tag(self, tag, values, *args, **kwargs):
        """Select the entries which songs all have one of
        the values for the given tag.
        """

        pattern_values = self.get_model().get_keys_by_tag(tag, values)
        self.set_selected(pattern_values, *args, **kwargs)

    def __get_selected_songs(self, sort=False):
        model, paths = self.get_selection().get_selected_rows()
        songs = model.get_songs(paths)
        if sort:
            return sorted(songs, key=operator.attrgetter("sort_key"))
        return songs



# Copyright 2016 Ryan Dellenbaugh
#           2017 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import random
import time

from quodlibet import _, print_d, print_w
from quodlibet.plugins.query import QueryPlugin, QueryPluginError, markup_for_syntax


class PythonQuery(QueryPlugin):
    PLUGIN_ID = "python_query"
    PLUGIN_NAME = _("Python Query")
    PLUGIN_DESC = _("🐍Use Python expressions in queries.")
    key = "python"
    query_syntax = _("@(python: expression)")
    query_description = _(
        "The variable <tt>s</tt> (or <tt>a</tt>) is the song / album being "
        "matched."
        "\n\n"
        "<tt>_ts</tt> is a (real number) timestamp at start of query."
        "\n\n"
        "Modules <tt>time</tt> and <tt>random</tt> are also available, and the "
        "class <tt>Random</tt> (==<tt>random.Random</tt>) too."
    )
    usage = markup_for_syntax(query_syntax) + "\n\n" + query_description

    def __init__(self):
        print_d("Initialising")
        self._globals = {"random": random, "Random": random.Random, "time": time}
        self._reported = set()
        self._raw_body = None

    def search(self, data, body):
        try:
            self._globals["s"] = data
            # Albums can be queried too...
            self._globals["a"] = data
            # eval modifies the globals in place, it seems
            return eval(body, dict(self._globals))
        except Exception as e:
            key = str(e)
            if key not in self._reported:
                self._reported.add(key)
                print_w(
                    f"{type(e).__name__}({key}) in expression {self._raw_body!r}. "
                    f"Example failing data: {self._globals}"
                )
            return False

    def parse_body(self, body):
        if body is None:
            raise QueryPluginError
        self._raw_body = body.strip()
        self._reported.clear()
        try:
            self._globals.update(_ts=time.time())
            return compile(body.strip(), "query", "eval")
        except SyntaxError as e:
            print_w(f"Couldn't compile query ({e})")
            raise QueryPluginError from e



# Copyright 2005 Joe Wreschnig
#      2017-2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys
import threading
import time
from urllib.request import urlopen, Request

from gi.repository import Gtk, GLib, Pango, Gdk
import feedparser

import quodlibet
from quodlibet import _
from quodlibet import config
from quodlibet import formats
from quodlibet import print_d
from quodlibet import qltk
from quodlibet import util
from quodlibet import app

from quodlibet.browsers import Browser
from quodlibet.formats import AudioFile
from quodlibet.formats.remote import RemoteFile
from quodlibet.qltk.getstring import GetStringDialog
from quodlibet.qltk.msg import ErrorMessage
from quodlibet.qltk.songsmenu import SongsMenu
from quodlibet.qltk.views import AllTreeView
from quodlibet.qltk import Icons
from quodlibet.util import connect_obj, print_w
from quodlibet.qltk.x import ScrolledWindow, Align, Button, MenuItem
from quodlibet.util.path import uri_is_valid
from quodlibet.util.picklehelper import pickle_load, pickle_dump, PickleError


FEEDS = os.path.join(quodlibet.get_user_dir(), "feeds")
DND_URI_LIST, DND_MOZ_URL = range(2)

# Migration path for pickle
sys.modules["browsers.audiofeeds"] = sys.modules[__name__]


class InvalidFeedError(ValueError):
    pass


class Feed(list):
    def __init__(self, uri):
        self.name = _("Unknown")
        self.uri = uri
        self.changed = False
        self.website = ""
        self.__lastgot = 0

    def get_age(self):
        return time.time() - self.__lastgot

    @staticmethod
    def __fill_af(feed, af):
        try:
            af["title"] = feed.title or _("Unknown")
        except (TypeError, AttributeError):
            af["title"] = _("Unknown")
        try:
            af["date"] = "%04d-%02d-%02d" % feed.modified_parsed[:3]
        except (AttributeError, TypeError):
            pass

        for songkey, feedkey in [
            ("website", "link"),
            ("description", "tagline"),
            ("language", "language"),
            ("copyright", "copyright"),
            ("organization", "publisher"),
            ("license", "license"),
        ]:
            try:
                value = getattr(feed, feedkey)
            except AttributeError:
                pass
            else:
                if value and value not in af.list(songkey):
                    af.add(songkey, value)

        try:
            af.add("artwork_url", feed.image["href"])
        except (AttributeError, KeyError):
            pass
        try:
            author = feed.author_detail
        except AttributeError:
            try:
                author = feed.author
            except AttributeError:
                pass
            else:
                if author and author not in af.list("artist"):
                    af.add("artist", author)
        else:
            try:
                if author.email and author.email not in af.list("contact"):
                    af.add("contact", author.email)
            except AttributeError:
                pass
            try:
                if author.name and author.name not in af.list("artist"):
                    af.add("artist", author.name)
            except AttributeError:
                pass

        try:
            values = feed.contributors
        except AttributeError:
            pass
        else:
            for value in values:
                try:
                    value = value.name
                except AttributeError:
                    pass
                else:
                    if value and value not in af.list("performer"):
                        af.add("performer", value)

        try:
            af["~#length"] = util.parse_time(feed.itunes_duration)
        except (AttributeError, ValueError):
            pass

        try:
            values = dict(feed.categories).values()
        except AttributeError:
            pass
        else:
            for value in values:
                if value and value not in af.list("genre"):
                    af.add("genre", value)

    def parse(self):
        req = Request(self.uri, method="HEAD")
        try:
            with urlopen(req, timeout=5) as head:
                # Some requests don't support status, e.g. file://
                if hasattr(head, "status"):
                    print_d(
                        f"Feed URL {self.uri!r} ({head.url}) "
                        f"returned HTTP {head.status}, "
                        f"with content {head.headers.get('Content-Type')}"
                    )
                    if head.status and head.status >= 400:
                        return False
                if head.headers.get("Content-Type").lower().startswith("audio"):
                    print_w("Looks like an audio stream / radio, not a audio feed.")
                    return False
            # Don't pass feedparser URLs
            # see https://github.com/kurtmckee/feedparser/pull/80#issuecomment-449543486
            content = urlopen(self.uri, timeout=15).read()
        except OSError as e:
            print_w(f"Couldn't fetch content from {self.uri} ({e})")
            return False
        try:
            doc = feedparser.parse(content)
        except Exception as e:
            print_w(f"Couldn't parse feed: {self.uri} ({e})")
            return False

        try:
            album = doc.channel.title
        except AttributeError:
            print_w(f"No channel title in {doc}")
            return False

        if album:
            self.name = album
        else:
            self.name = _("Unknown")

        defaults = AudioFile({"feed": self.uri})
        try:
            self.__fill_af(doc.channel, defaults)
        except Exception as e:
            print_w(f"Error creating feed data: {self.uri} ({e})")
            return False

        entries = []
        uris = set()
        print_d("Found %d entries in channel" % len(doc.entries))
        for entry in doc.entries:
            try:
                for enclosure in entry.enclosures:
                    try:
                        if (
                            "audio" in enclosure.type
                            or "ogg" in enclosure.type
                            or formats.filter(enclosure.url)
                        ):
                            uri = enclosure.url
                            if not isinstance(uri, str):
                                uri = uri.decode("utf-8")
                            try:
                                size = float(enclosure.length)
                            except (AttributeError, ValueError):
                                size = 0
                            entries.append((uri, entry, size))
                            uris.add(uri)
                            break
                    except AttributeError:
                        pass
            except AttributeError:
                print_d(f"No enclosures found in {entry}")

        for entry in list(self):
            if entry["~uri"] not in uris:
                self.remove(entry)
            else:
                uris.remove(entry["~uri"])
        print_d("Successfully got %d episodes in channel" % len(entries))
        entries.reverse()
        for uri, entry, size in entries:
            if uri in uris:
                song = RemoteFile(uri)
                song["~#size"] = size
                song.fill_metadata = False
                song.update(defaults)
                song["album"] = self.name
                try:
                    self.__fill_af(entry, song)
                except Exception as e:
                    print_d(f"Couldn't convert {uri} to AudioFile ({e})")
                else:
                    self.insert(0, song)
        self.__lastgot = time.time()
        return bool(uris)


class AddFeedDialog(GetStringDialog):
    def __init__(self, parent):
        super().__init__(
            qltk.get_top_parent(parent),
            _("New Feed"),
            _("Enter the podcast / audio feed location:"),
            button_label=_("_Add"),
            button_icon=Icons.LIST_ADD,
        )

    def run(self, text="", clipboard=True, test=False):
        uri = super().run(text=text, clipboard=clipboard, test=test)
        if uri:
            if not isinstance(uri, str):
                uri = uri.decode("utf-8")
            return Feed(uri)
        return None

    def _verify_clipboard(self, text):
        # try to extract a URI from the clipboard
        for line in text.splitlines():
            line = line.strip()

            if uri_is_valid(line):
                return line
        return None


def hacky_py2_unpickle_recover(fileobj):
    """Can raise anything"""

    # We just recover the uri and let it refresh the feed later

    def lookup_func(func, mod, name):
        class Feed(list):
            pass

        class Bar(dict):
            pass

        if name == "Feed":
            return Feed
        return Bar

    with open(FEEDS, "rb") as fileobj:
        feeds = pickle_load(fileobj, lookup_func)

    uris = []
    for item in feeds:
        d = item.__dict__
        if b"uri" in d:
            v = d[b"uri"]
        elif "uri" in d:
            v = d["uri"]
        else:
            continue
        if isinstance(v, bytes):
            v = v.decode("utf-8")
        uris.append(v)

    return [Feed(u) for u in uris]


class Podcasts(Browser):
    """
    Allows interacting with remote feeds for playing and exploring podcasts
    and their episodes (or other forms of audio feeds)
    Formerly known as the AudioFeeds browser.
    """

    __feeds = Gtk.ListStore(object)  # unread

    headers = (
        "title artist performer ~people album date website language "
        "copyright organization license contact"
    ).split()

    name = _("Podcasts")
    accelerated_name = _("_Podcasts")
    keys = ["AudioFeeds", "Podcasts"]
    priority = 20
    uses_main_library = False

    def pack(self, songpane):
        container = qltk.ConfigRHPaned("browsers", "audiofeeds_pos", 0.4)
        self.show()
        container.pack1(self, True, False)
        container.pack2(songpane, True, False)
        return container

    def unpack(self, container, songpane):
        container.remove(songpane)
        container.remove(self)

    @staticmethod
    def cell_data(col, render, model, iter, data):
        if model[iter][0].changed:
            render.markup = util.bold(model[iter][0].name)
        else:
            render.markup = util.escape(model[iter][0].name)
        render.set_property("markup", render.markup)

    @classmethod
    def changed(cls, feeds):
        for row in cls.__feeds:
            if row[0] in feeds:
                row[0].changed = True
                row[0] = row[0]
        Podcasts.write()

    @classmethod
    def write(cls):
        feeds = [row[0] for row in cls.__feeds]
        with open(FEEDS, "wb") as f:
            pickle_dump(feeds, f, 2)

    @classmethod
    def init(cls, library):
        uris = set()
        feeds = []

        try:
            with open(FEEDS, "rb") as fileobj:
                feeds = pickle_load(fileobj)
        except (OSError, PickleError):
            try:
                with open(FEEDS, "rb") as fileobj:
                    feeds = hacky_py2_unpickle_recover(fileobj)
            except Exception:
                pass

        for feed in feeds:
            if feed.uri in uris:
                continue
            cls.__feeds.append(row=[feed])
            uris.add(feed.uri)

        GLib.idle_add(cls.__do_check)

    @classmethod
    def reload(cls, library):
        cls.__feeds = Gtk.ListStore(object)  # unread
        cls.init(library)

    @classmethod
    def __do_check(cls):
        thread = threading.Thread(target=cls.__check, args=(), daemon=True)
        thread.start()

    @classmethod
    def __check(cls):
        for row in cls.__feeds:
            feed = row[0]
            if feed.get_age() < 2 * 60 * 60:
                continue
            if feed.parse():
                feed.changed = True
                row[0] = feed
        cls.write()
        GLib.timeout_add(60 * 60 * 1000, cls.__do_check)

    def __init__(self, library):
        super().__init__(spacing=6)
        self.set_orientation(Gtk.Orientation.VERTICAL)

        self._view = view = AllTreeView()
        self.__render = render = Gtk.CellRendererText()
        render.set_property("ellipsize", Pango.EllipsizeMode.END)
        col = Gtk.TreeViewColumn("Audio Feeds", render)
        col.set_cell_data_func(render, Podcasts.cell_data)
        view.append_column(col)
        view.set_model(self.__feeds)
        view.set_rules_hint(True)
        view.set_headers_visible(False)
        swin = ScrolledWindow()
        swin.set_shadow_type(Gtk.ShadowType.IN)
        swin.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        swin.add(view)
        self.pack_start(swin, True, True, 0)

        new = Button(_("_Add Feed…"), Icons.LIST_ADD, Gtk.IconSize.MENU)
        new.connect("clicked", self.__new_feed)
        view.get_selection().connect("changed", self.__changed)
        view.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        view.connect("popup-menu", self._popup_menu)

        targets = [
            ("text/uri-list", 0, DND_URI_LIST),
            ("text/x-moz-url", 0, DND_MOZ_URL),
        ]
        targets = [Gtk.TargetEntry.new(*t) for t in targets]

        view.drag_dest_set(Gtk.DestDefaults.ALL, targets, Gdk.DragAction.COPY)
        view.connect("drag-data-received", self.__drag_data_received)
        view.connect("drag-motion", self.__drag_motion)
        view.connect("drag-leave", self.__drag_leave)

        connect_obj(self, "destroy", self.__save, view)

        self.pack_start(Align(new, left=3, bottom=3), False, True, 0)

        for child in self.get_children():
            child.show_all()

    def menu(self, songs, library, items):
        return SongsMenu(library, songs, download=True, items=items)

    def __drag_motion(self, view, ctx, x, y, time):
        targets = [t.name() for t in ctx.list_targets()]
        if "text/x-quodlibet-songs" not in targets:
            view.get_parent().drag_highlight()
            return True
        return False

    def __drag_leave(self, view, ctx, time):
        view.get_parent().drag_unhighlight()

    def __drag_data_received(self, view, ctx, x, y, sel, tid, etime):
        view.emit_stop_by_name("drag-data-received")
        targets = [
            ("text/uri-list", 0, DND_URI_LIST),
            ("text/x-moz-url", 0, DND_MOZ_URL),
        ]
        targets = [Gtk.TargetEntry.new(*t) for t in targets]

        view.drag_dest_set(Gtk.DestDefaults.ALL, targets, Gdk.DragAction.COPY)
        if tid == DND_URI_LIST:
            uri = sel.get_uris()[0]
        elif tid == DND_MOZ_URL:
            uri = sel.data.decode("utf16", "replace").split("\n")[0]
        else:
            ctx.finish(False, False, etime)
            return

        ctx.finish(True, False, etime)

        feed = Feed(uri.encode("ascii", "replace"))
        feed.changed = feed.parse()
        if feed:
            self.__feeds.append(row=[feed])
            Podcasts.write()
        else:
            self.feed_error(feed).run()

    def _popup_menu(self, view: Gtk.Widget) -> Gtk.Menu | None:
        model, paths = self._view.get_selection().get_selected_rows()
        menu = Gtk.Menu()
        refresh = MenuItem(
            _("_Refresh"),
            Icons.VIEW_REFRESH,
            tooltip=_("Search source for new episodes"),
        )
        rebuild = MenuItem(
            _("_Rebuild"),
            Icons.EDIT_FIND_REPLACE,
            tooltip=_("Remove all existing episodes then reload from source"),
        )
        delete = MenuItem(
            _("_Delete"),
            Icons.EDIT_DELETE,
            tooltip=_("Remove this podcast and its episodes"),
        )

        connect_obj(refresh, "activate", self.__refresh, [model[p][0] for p in paths])
        connect_obj(rebuild, "activate", self.__rebuild, [model[p][0] for p in paths])
        connect_obj(delete, "activate", self.__remove_paths, model, paths)

        menu.append(refresh)
        menu.append(rebuild)
        menu.append(delete)
        menu.show_all()
        menu.connect("selection-done", lambda m: m.destroy())

        if self._view.popup_menu(menu, 0, Gtk.get_current_event_time()):
            return menu
        return None

    def __save(self, view):
        Podcasts.write()

    def __refresh(self, feeds):
        changed = list(filter(Feed.parse, feeds))
        Podcasts.changed(changed)

    def __rebuild(self, feeds):
        for feed in feeds:
            feed.clear()
        changed = list(filter(Feed.parse, feeds))
        Podcasts.changed(changed)

    def __remove_paths(self, model, paths):
        for path in paths:
            model.remove(model.get_iter(path))

    def activate(self):
        self.__changed(self._view.get_selection())

    def __changed(self, selection):
        model, paths = selection.get_selected_rows()
        if model and paths:
            songs = []
            for path in paths:
                model[path][0].changed = False
                songs.extend(model[path][0])
            self.songs_selected(songs, True)
            config.set(
                "browsers",
                "audiofeeds",
                "\t".join([model[path][0].name for path in paths]),
            )

    def __new_feed(self, activator):
        feed = AddFeedDialog(self).run()
        if feed is not None:
            feed.changed = feed.parse()
            if feed:
                self.__feeds.append(row=[feed])
                Podcasts.write()
            else:
                self.feed_error(feed).run()

    def feed_error(self, feed: Feed) -> ErrorMessage:
        return ErrorMessage(
            self,
            _("Unable to add feed"),
            _(
                "%s could not be added. The server may be down, "
                "or the location may not be a podcast / audio feed."
            )
            % util.bold(util.escape(feed.uri)),
            escape_desc=False,
        )

    def restore(self):
        try:
            names = config.get("browsers", "audiofeeds").split("\t")
        except Exception:
            pass
        else:
            self._view.select_by_func(lambda r: r[0].name in names)


browsers = []
if not app.player or app.player.can_play_uri("http://"):
    browsers = [Podcasts]
else:
    print_w(
        _(
            "The current audio backend does not support URLs, "
            "Podcast browser disabled."
        )
    )



# Copyright 2005 Joe Wreschnig, Michael Urman
#       2021-22 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


from gi.repository import Gtk

from quodlibet import _
from quodlibet import util
from quodlibet.qltk import get_top_parent
from quodlibet.qltk.icons import Icons
from quodlibet.qltk.window import Dialog
from quodlibet.util import escape
from senf import fsn2text, path2fsn


class Message(Gtk.MessageDialog, Dialog):
    """A message dialog that destroys itself after it is run, uses
    markup, and defaults to an 'OK' button."""

    def __init__(
        self,
        kind: type,
        parent: Gtk.Widget,
        title: str,
        description: str,
        buttons: Gtk.ButtonsType = Gtk.ButtonsType.OK,
        escape_desc: bool = True,
    ):
        parent = get_top_parent(parent)
        markup = (
            f"<span weight='bold' size='larger'>{escape(title)}</span>\n\n"
            + escape(description)
            if escape_desc
            else description
        )
        super().__init__(
            transient_for=parent,
            modal=True,
            destroy_with_parent=True,
            message_type=kind,
            buttons=buttons,
        )
        self.set_markup(markup)

    def run(self, destroy=True):
        resp = super().run()
        if destroy:
            self.destroy()
        return resp


class CancelRevertSave(Gtk.MessageDialog, Dialog):
    def __init__(self, parent):
        title = _("Discard tag changes?")
        description = _(
            "Tags have been changed but not saved. Save these "
            "files, or revert and discard changes?"
        )
        text = f"<span weight='bold' size='larger'>{title}</span>\n\n{description}"
        parent = get_top_parent(parent)
        super().__init__(
            transient_for=parent,
            flags=0,
            message_type=Gtk.MessageType.WARNING,
            buttons=Gtk.ButtonsType.NONE,
        )

        self.add_icon_button(_("_Save"), Icons.DOCUMENT_SAVE, Gtk.ResponseType.YES)
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_icon_button(_("_Revert"), Icons.DOCUMENT_REVERT, Gtk.ResponseType.NO)

        self.set_default_response(Gtk.ResponseType.NO)
        self.set_markup(text)

    def run(self):
        resp = super().run()
        self.destroy()
        return resp


class ErrorMessage(Message):
    """Like Message, but uses an error-indicating picture."""

    def __init__(self, *args, **kwargs):
        super().__init__(Gtk.MessageType.ERROR, *args, **kwargs)


class WarningMessage(Message):
    """Like Message, but uses a warning-indicating picture."""

    def __init__(self, *args, **kwargs):
        super().__init__(Gtk.MessageType.WARNING, *args, **kwargs)


class ConfirmationPrompt(WarningMessage):
    """Dialog to confirm actions, given a parent, title, description, and
    OK-button text"""

    RESPONSE_INVOKE = Gtk.ResponseType.YES

    def __init__(
        self,
        parent,
        title,
        description,
        ok_button_text,
        ok_button_icon=Icons.SYSTEM_RUN,
    ):
        super().__init__(parent, title, description, buttons=Gtk.ButtonsType.NONE)

        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_icon_button(ok_button_text, ok_button_icon, self.RESPONSE_INVOKE)
        self.set_default_response(Gtk.ResponseType.CANCEL)


class ConfirmFileReplace(WarningMessage):
    RESPONSE_REPLACE = 1

    def __init__(self, parent, path):
        title = _("File exists")
        fn_format = util.bold(fsn2text(path2fsn(path)))
        description = _("Replace %(file-name)s?") % {"file-name": fn_format}

        super().__init__(parent, title, description, buttons=Gtk.ButtonsType.NONE)

        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_icon_button(
            _("_Replace File"), Icons.DOCUMENT_SAVE, self.RESPONSE_REPLACE
        )
        self.set_default_response(Gtk.ResponseType.CANCEL)



# Copyright 2012-2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
from collections.abc import Iterable

from gi.repository import Gtk
from gi.repository import Pango

from quodlibet import _
from quodlibet import qltk, util
from quodlibet.qltk.entry import UndoEntry, ValidatingEntry
from quodlibet.qltk.views import RCMHintedTreeView, HintedTreeView
from quodlibet.qltk.x import MenuItem, Button, Align
from quodlibet.qltk import Icons
from quodlibet.query import Query
from quodlibet.util.json_data import JSONObjectDict
from quodlibet.util import connect_obj
from quodlibet.qltk.getstring import GetStringDialog


class JSONBasedEditor(qltk.UniqueWindow):
    """
    Flexible editor for objects extending `JSONObject`
    (held in a `JSONObjectDict`)
    TODO: validation, especially for name.
    """

    _WIDTH = 800
    _HEIGHT = 400

    def __init__(self, proto_cls, values, filename, title):
        if self.is_not_unique():
            return
        super().__init__()
        self.proto_cls = proto_cls
        self.current = None
        self.filename = filename
        self.name = proto_cls.NAME or proto_cls.__name__
        self.input_entries = {}
        self.set_border_width(12)
        self.set_title(title)
        self.set_default_size(self._WIDTH, self._HEIGHT)

        self.add(Gtk.HBox(spacing=6))
        self.get_child().set_homogeneous(True)
        self.accels = Gtk.AccelGroup()

        # Set up the model for this widget
        self.model = Gtk.ListStore(object)
        self._fill_values(values)

        # The browser for existing data
        self.view = view = RCMHintedTreeView(model=self.model)
        view.set_headers_visible(False)
        view.set_reorderable(True)
        view.set_rules_hint(True)
        render = Gtk.CellRendererText()
        render.set_padding(3, 6)
        render.props.ellipsize = Pango.EllipsizeMode.END
        column = Gtk.TreeViewColumn("", render)
        column.set_cell_data_func(render, self.__cdf)
        view.append_column(column)
        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        sw.add(view)
        self.get_child().pack_start(sw, True, True, 0)

        vbox = Gtk.VBox(spacing=6)
        # Input for new ones.
        frame = self.__build_input_frame()
        vbox.pack_start(frame, False, True, 0)

        # Add context menu
        menu = Gtk.Menu()
        rem = MenuItem(_("_Remove"), Icons.LIST_REMOVE)
        keyval, mod = Gtk.accelerator_parse("Delete")
        rem.add_accelerator(
            "activate", self.accels, keyval, mod, Gtk.AccelFlags.VISIBLE
        )
        connect_obj(rem, "activate", self.__remove, view)
        menu.append(rem)
        menu.show_all()
        view.connect("popup-menu", self.__popup, menu)
        view.connect("key-press-event", self.__view_key_press)
        connect_obj(self, "destroy", Gtk.Menu.destroy, menu)

        # New and Close buttons
        bbox = Gtk.HButtonBox()
        self.remove_but = Button(_("_Remove"), Icons.LIST_REMOVE)
        self.remove_but.set_sensitive(False)
        self.new_but = Button(_("_New"), Icons.DOCUMENT_NEW)
        self.new_but.connect("clicked", self._new_item)
        bbox.pack_start(self.new_but, True, True, 0)
        close = Button(_("_Close"), Icons.WINDOW_CLOSE)
        connect_obj(close, "clicked", qltk.Window.destroy, self)
        bbox.pack_start(close, True, True, 0)
        vbox.pack_end(bbox, False, True, 0)

        self.get_child().pack_start(vbox, True, True, 0)
        # Initialise
        self.selection = view.get_selection()

        self.selection.connect("changed", self.__select)
        self.connect("destroy", self.__finish)
        self.get_child().show_all()

    def _find(self, name):
        for row in self.model:
            if row[0].name == name:
                return row[0]
        return None

    def _new_item(self, button):
        # Translators: New Command/Entry/Item/...
        current_name = name = _("New %s") % self.name
        n = 2
        while True:
            if self._find(current_name):
                current_name = "%s (%d)" % (name, n)
                n += 1
                continue
            break
        self.model.append(row=(self.proto_cls(name=current_name),))

    def _new_widget(self, key, val):
        """
        Creates a Gtk.Entry subclass
        appropriate for a field named `key` with value `val`
        """
        callback = signal = None
        if isinstance(val, bool):
            entry = Gtk.CheckButton()
            callback = self.__toggled_widget
            signal = "toggled"
        elif isinstance(val, int):
            adj = Gtk.Adjustment.new(0, 0, 9999, 1, 10, 0)
            entry = Gtk.SpinButton(adjustment=adj)
            entry.set_numeric(True)
            callback = self.__changed_numeric_widget
        elif "pattern" in key:
            entry = ValidatingEntry(validator=Query.validator)
        else:
            entry = UndoEntry()
        entry.connect(signal or "changed", callback or self.__changed_widget, key)
        return entry

    def __refresh_view(self):
        model, iter = self.selection.get_selected()
        self.model.emit("row-changed", model[iter].path, iter)

    def __changed_widget(self, entry, key):
        if self.current:
            setattr(self.current, key, str(entry.get_text()))
            self.__refresh_view()

    def __changed_numeric_widget(self, entry, key):
        if self.current:
            setattr(self.current, key, int(entry.get_text() or 0))
            self.__refresh_view()

    def __toggled_widget(self, entry, key):
        if self.current:
            setattr(self.current, key, bool(entry.get_active()))
            self.__refresh_view()

    def _populate_fields(self, obj):
        """Populates the input fields based on the `JSONData` object `obj`"""
        for fn, val in obj.data:
            widget = self.input_entries[fn]
            widget.set_sensitive(True)
            # TODO: link this logic better with the creational stuff
            if isinstance(val, bool):
                widget.set_active(val)
            elif isinstance(val, int):
                widget.set_value(int(val))
            elif isinstance(val, str):
                widget.set_text(val or "")

    def __build_input_frame(self):
        t = Gtk.Table(n_rows=2, n_columns=3)
        t.set_row_spacings(6)
        t.set_col_spacing(0, 3)
        t.set_col_spacing(1, 12)

        empty = self.proto_cls("empty")
        for i, (key, val) in enumerate(empty.data):
            field = empty.field(key)
            field_name = self.get_field_name(field, key)
            l = Gtk.Label(label=field_name + ":")
            entry = self._new_widget(key, val)
            entry.set_sensitive(False)
            if field.doc:
                entry.set_tooltip_text(field.doc)
            # Store these away in a map for later access
            self.input_entries[key] = entry
            l.set_mnemonic_widget(entry)
            l.set_use_underline(True)
            l.set_alignment(0.0, 0.5)
            if isinstance(val, int | bool):
                align = Align(entry, halign=Gtk.Align.START)
                t.attach(align, 1, 2, i, i + 1)
            else:
                t.attach(entry, 1, 2, i, i + 1)
            t.attach(l, 0, 1, i, i + 1, xoptions=Gtk.AttachOptions.FILL)
        frame = qltk.Frame(label=self.name, child=t)
        self.input_entries["name"].grab_focus()
        return frame

    @staticmethod
    def get_field_name(field, key):
        field_name = field.human_name or (key and key.replace("_", " "))
        return field_name and util.capitalize(field_name) or _("(unknown)")

    def _fill_values(self, data):
        if not data:
            return
        for _name, obj in data.items():
            self.model.append(row=[obj])

    def _update_current(self, new_selection=None):
        if new_selection:
            self.selection = new_selection
        model, iter = self.selection.get_selected()
        if iter:
            self.current = model[iter][0]

    def __select(self, selection):
        self._update_current(selection)
        self.remove_but.set_sensitive(bool(iter))
        if iter is not None:
            self._populate_fields(self.current)

    def __remove(self, view):
        view.remove_selection()

    def __popup(self, view, menu):
        return view.popup_menu(menu, 0, Gtk.get_current_event_time())

    def __view_key_press(self, view, event):
        if event.keyval == Gtk.accelerator_parse("Delete")[0]:
            self.__remove(view)

    def __cdf(self, column, cell, model, iter, data):
        row = model[iter]
        obj = row[0]
        obj_name = util.escape(obj.name)
        obj_description = util.escape(str(obj))
        markup = f"{util.bold(obj_name)}\n{obj_description}"
        cell.markup = markup
        cell.set_property("markup", markup)

    def __finish(self, widget):
        # TODO: Warn about impending deletion of nameless items, or something
        all = JSONObjectDict.from_list([row[0] for row in self.model if row[0].name])
        all.save(filename=self.filename)


class TagListEditor(qltk.Window):
    """Dialog to edit a list of tag names."""

    _WIDTH = 600
    _HEIGHT = 300

    def __init__(self, title, values=None):
        super().__init__()
        self.use_header_bar()
        self.set_border_width(12)
        self.set_title(title)
        self.set_default_size(self._WIDTH, self._HEIGHT)

        vbox = Gtk.VBox(spacing=12)
        hbox = Gtk.HBox(spacing=12)

        # Set up the model for this widget
        self.model = Gtk.ListStore(str)
        self.__fill_values(values or [])

        def on_row_activated(view, path, column):
            self._renderer.set_property("editable", True)
            view.set_cursor(path, view.get_columns()[0], start_editing=True)

        # Main view
        view = self.view = HintedTreeView(model=self.model)
        view.set_fixed_height_mode(True)
        view.set_headers_visible(False)
        view.connect("row-activated", on_row_activated)

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.add(view)
        sw.set_size_request(-1, max(sw.size_request().height, 100))
        hbox.pack_start(sw, True, True, 0)

        self.__setup_column(view)

        # Context menu
        menu = Gtk.Menu()
        remove_item = MenuItem(_("_Remove"), Icons.LIST_REMOVE)
        menu.append(remove_item)
        menu.show_all()
        view.connect("popup-menu", self.__popup, menu)
        connect_obj(remove_item, "activate", self.__remove, view)

        # Add and Remove buttons
        vbbox = Gtk.VButtonBox()
        vbbox.set_layout(Gtk.ButtonBoxStyle.START)
        vbbox.set_spacing(6)
        add = Button(_("_Add…"), Icons.LIST_ADD)
        add.connect("clicked", self.__add)
        vbbox.pack_start(add, False, True, 0)
        remove = Button(_("_Remove"), Icons.LIST_REMOVE)
        remove.connect("clicked", self.__remove)
        vbbox.pack_start(remove, False, True, 0)
        edit = Button(_("_Edit"), Icons.LIST_EDIT)
        edit.connect("clicked", self.__edit)
        vbbox.pack_start(edit, False, True, 0)
        hbox.pack_start(vbbox, False, True, 0)
        vbox.pack_start(hbox, True, True, 0)

        # Close buttons
        bbox = Gtk.HButtonBox()
        self.remove_but = Button(_("_Remove"), Icons.LIST_REMOVE)
        self.remove_but.set_sensitive(False)
        close = Button(_("_Close"), Icons.WINDOW_CLOSE)
        connect_obj(close, "clicked", qltk.Window.destroy, self)
        bbox.set_layout(Gtk.ButtonBoxStyle.END)
        if not self.has_close_button():
            bbox.pack_start(close, True, True, 0)
            vbox.pack_start(bbox, False, True, 0)

        # Finish up
        self.add(vbox)
        self.get_child().show_all()

    def __start_editing(self, _render, editable, path):
        editable.set_text(self.model[path][0])

    def __edited(self, _render, path, new_name):
        self.model[path][0] = new_name
        self.model.row_changed(path, self.model.get_iter(path))

    def __setup_column(self, view):
        def tag_cdf(column, cell, model, iter, data):
            row = model[iter]
            if row:
                cell.set_property("text", row[0])

        def desc_cdf(column, cell, model, iter, data):
            row = model[iter]
            if row:
                cell.set_property("markup", util.italic(util.tag(row[0])))

        def __create_cell_renderer():
            r = Gtk.CellRendererText()
            r.connect("editing-started", self.__start_editing)
            r.connect("edited", self.__edited)
            return r

        self._renderer = renderer = __create_cell_renderer()
        column = Gtk.TreeViewColumn(_("Tag expression"), renderer)
        column.set_cell_data_func(renderer, tag_cdf)
        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        column.set_expand(True)
        view.append_column(column)

        renderer = Gtk.CellRendererText()
        renderer.set_property("ellipsize", Pango.EllipsizeMode.END)
        renderer.set_property("sensitive", False)
        column = Gtk.TreeViewColumn(_("Description"), renderer)
        column.set_cell_data_func(renderer, desc_cdf)
        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        column.set_expand(True)

        view.append_column(column)
        view.set_headers_visible(True)

    def __fill_values(self, data: Iterable[str]):
        for s in data:
            self.model.append(row=[s])

    @property
    def tags(self):
        """Returns the tag names as edited"""
        return [row[0] for row in self.model if row]

    def __remove(self, *args):
        self.view.remove_selection()

    def __add(self, *args):
        tooltip = _("Tag expression e.g. people:real or ~album~year")
        dialog = GetStringDialog(
            self, _("Enter new tag"), "", button_icon=None, tooltip=tooltip
        )
        new = dialog.run()
        if new:
            self.model.append(row=[new])

    def __edit(self, *args):
        path, col = self.view.get_cursor()
        tooltip = _("Tag expression e.g. people:real or ~album~year")
        dialog = GetStringDialog(
            self, _("Edit tag expression"), "", button_icon=None, tooltip=tooltip
        )
        edited = dialog.run(text=self.model[path][0])
        if edited:
            self.model[path][0] = edited

    def __popup(self, view, menu):
        return view.popup_menu(menu, 0, Gtk.get_current_event_time()).show()



# Copyright 2012 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from mutagen.monkeysaudio import MonkeysAudio

from ._audio import translate_errors
from ._apev2 import APEv2File


class MonkeysAudioFile(APEv2File):
    format = "Monkey's Audio"

    def __init__(self, filename):
        with translate_errors():
            audio = MonkeysAudio(filename)
        super().__init__(filename, audio)
        self["~#length"] = int(audio.info.length)
        self["~#channels"] = audio.info.channels
        self["~#samplerate"] = audio.info.sample_rate
        if hasattr(audio.info, "bits_per_sample"):
            self["~#bitdepth"] = audio.info.bits_per_sample
        self.sanitize(filename)


loader = MonkeysAudioFile
types = [MonkeysAudioFile]
extensions = [".ape"]



# Copyright 2016 Ryan Dellenbaugh
#           2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from typing import Any

from gi.repository import Gtk

from quodlibet import _
from quodlibet import util
from quodlibet.formats import AudioFile
from quodlibet.plugins import PluginHandler, PluginManager
from quodlibet.qltk import Icons, Align, Frame


def markup_for_syntax(text: str) -> str:
    return f"<b><tt>{util.escape(text)}</tt></b>"


class QueryPlugin:
    """
    Query plugins provide extensions to the search query syntax used in '@'
    queries, like '@(name: body)'

    The plugin must provide the search instance method:

        self.search(song, body)

    The plugin may optionally provide the parse_body method:

        self.parse_body(body)

    This method will be called once, when the query is parsed, with the body
    string from the query, or None if the query contained no body. It should
    return a value representing the parsed body, to be used in self.search.

    If the provided body is invalid, the method may raise a QueryPluginError
    to indicate that all matches should fail. In this case search will not be
    called.
    """

    def search(self, song: AudioFile, body: Any | None) -> bool:
        """Whether to match the given song
        :param song: the song in question
        :param body: the query body
        :returns:  whether the song matches the query for the given body.
        """
        raise NotImplementedError

    key: str | None = None
    """The name used for query syntax, if not the name of the plugin"""

    PLUGIN_ICON = Icons.EDIT_FIND

    usage: str | None = None
    """Override this markup to show example usage for users"""

    def parse_body(self, body: str) -> str:
        return body

    @classmethod
    def PluginPreferences(cls, window):  # noqa
        if not cls.usage:
            return Gtk.VBox()
        label = Gtk.Label(
            label=cls.usage,
            use_markup=True,
            xalign=0.0,
            yalign=0.5,
            wrap=True,
            width_chars=30,
            selectable=True,
        )
        return Frame(
            _("Usage"),
            child=Align(
                label, border=9, halign=Gtk.Align.START, valign=Gtk.Align.START
            ),
        )


class QueryPluginError(Exception):
    pass


class QueryPluginHandler(PluginHandler):
    """Maintains a dictionary of enabled Query Plugins by key"""

    def init_plugins(self):
        PluginManager.instance.register_handler(self)

    def __init__(self):
        self.plugins = {}

    def plugin_handle(self, plugin):
        return issubclass(plugin.cls, QueryPlugin)

    def plugin_enable(self, plugin):
        self.plugins[plugin.cls.key or plugin.name] = plugin.cls()

    def plugin_disable(self, plugin):
        del self.plugins[plugin.cls.key or plugin.name]

    def get_plugin(self, key):
        return self.plugins[key]


QUERY_HANDLER = QueryPluginHandler()



# Copyright 2012,2013 Christoph Reiter
#                2023 Nick Boultbee
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import sys
import os
from optparse import OptionParser

import quodlibet
from quodlibet import const
from quodlibet.util.dprint import print_

from .base import Command, CommandError
from . import commands

commands  # noqa


def _print_help(main_cmd, parser, file=None):
    """Print a short help list for all commands"""

    if file is None:
        file = sys.stdout

    parser.print_help(file=file)

    cl = ["", "Commands:"]
    for command in Command.COMMANDS:
        cl.append("   %-17s %s" % (command.NAME, command.DESCRIPTION))
    cl.append("")
    cl.append(
        f"See '{main_cmd} help <command>' for more information on a specific command."
    )

    print_("\n".join(cl), file=file)


def main(argv=None):
    if argv is None:
        argv = sys.argv

    quodlibet.init_cli()

    main_cmd = os.path.basename(argv[0])

    # the main optparser
    usage = f"{main_cmd} [--version] [--help] [--verbose] <command> [<args>]"
    parser = OptionParser(usage=usage)

    parser.remove_option("--help")
    parser.add_option("-h", "--help", action="store_true")
    parser.add_option("--version", action="store_true", help="print version")
    parser.add_option("-v", "--verbose", action="store_true", help="verbose output")

    # no args, print help (might change in the future)
    if len(argv) <= 1:
        _print_help(main_cmd, parser, file=sys.stderr)
        return 1

    # collect options for the main command and get the command offset
    offset = -1
    pre_command = []
    for i, a in enumerate(argv):
        if i == 0:
            continue
        if a.startswith("-"):
            pre_command.append(a)
        else:
            offset = i
            break

    # parse the global options
    options = parser.parse_args(pre_command)[0]

    # --help somewhere
    if options.help:
        _print_help(main_cmd, parser)
        return 0

    # --version somewhere
    if options.version:
        print_(f"{main_cmd} version {const.VERSION}")
        return 0

    # no sub command followed, help to stderr
    if offset == -1:
        _print_help(main_cmd, parser, file=sys.stderr)
        return 1
    arg = argv[offset]

    # special case help and list all commands
    if arg == "help":
        # no command, list all commands
        if len(argv) == 2:
            _print_help(main_cmd, parser)
            return 0

    # get the right sub command and pass the remaining args
    for command in Command.COMMANDS:
        if command.NAME == arg:
            cmd = command(main_cmd, options)
            try:
                cmd.execute(argv[offset + 1 :])
            except CommandError as e:
                print_(f"{command.NAME}: {e}", file=sys.stderr)
                return 1
            break
    else:
        print_(f"Unknown command '{arg}'. See '{main_cmd} help'.", file=sys.stderr)
        return 1

    return 0



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk
from senf import fsn2text

from quodlibet import qltk
from quodlibet import util
from quodlibet import _
from quodlibet.formats import AudioFileError
from quodlibet.qltk._editutils import OverwriteWarning, WriteFailedError
from quodlibet.qltk.views import HintedTreeView, TreeViewColumn
from quodlibet.qltk.wlw import WritingWindow
from quodlibet.qltk.x import Button, Align
from quodlibet.qltk.models import ObjectStore
from quodlibet.qltk import Icons
from quodlibet.util import connect_obj


class Entry:
    def __init__(self, song):
        self.song = song
        self.tracknumber = song("tracknumber")

    @property
    def name(self):
        return fsn2text(self.song("~basename"))


class TrackNumbers(Gtk.VBox):
    def __init__(self, prop, library):
        super().__init__(spacing=6)
        self.title = _("Track Numbers")
        self.set_border_width(12)

        label_start = Gtk.Label(label=_("Start fro_m:"), halign=Gtk.Align.END)
        label_start.set_use_underline(True)
        spin_start = Gtk.SpinButton()
        spin_start.set_range(0, 9999)
        spin_start.set_increments(1, 10)
        spin_start.set_value(1)
        label_start.set_mnemonic_widget(spin_start)

        label_total = Gtk.Label(label=_("_Total tracks:"), halign=Gtk.Align.END)
        label_total.set_use_underline(True)
        spin_total = Gtk.SpinButton()
        spin_total.set_range(0, 9999)
        spin_total.set_increments(1, 10)
        label_total.set_mnemonic_widget(spin_total)

        preview = qltk.Button(_("_Preview"), Icons.VIEW_REFRESH)

        grid = Gtk.Grid(row_spacing=4, column_spacing=4)
        grid.add(label_start)
        grid.attach_next_to(spin_start, label_start, Gtk.PositionType.RIGHT, 1, 1)
        grid.attach_next_to(label_total, label_start, Gtk.PositionType.BOTTOM, 1, 1)
        grid.attach_next_to(spin_total, label_total, Gtk.PositionType.RIGHT, 1, 1)
        grid.attach_next_to(
            Align(preview, halign=Gtk.Align.END),
            spin_start,
            Gtk.PositionType.RIGHT,
            1,
            1,
        )
        preview.props.hexpand = True

        model = ObjectStore()
        view = HintedTreeView(model=model)

        self.pack_start(grid, False, True, 0)

        render = Gtk.CellRendererText()
        column = TreeViewColumn(title=_("File"))
        column.pack_start(render, True)
        column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)

        def cell_data_file(column, cell, model, iter_, data):
            entry = model.get_value(iter_)
            cell.set_property("text", entry.name)

        column.set_cell_data_func(render, cell_data_file)

        view.append_column(column)
        render = Gtk.CellRendererText()
        render.set_property("editable", True)
        column = TreeViewColumn(title=_("Track"))
        column.pack_start(render, True)
        column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)

        def cell_data_track(column, cell, model, iter_, data):
            entry = model.get_value(iter_)
            cell.set_property("text", entry.tracknumber)

        column.set_cell_data_func(render, cell_data_track)

        view.append_column(column)
        view.set_reorderable(True)
        w = Gtk.ScrolledWindow()
        w.set_shadow_type(Gtk.ShadowType.IN)
        w.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        w.add(view)
        self.pack_start(w, True, True, 0)

        bbox = Gtk.HButtonBox()
        bbox.set_spacing(6)
        bbox.set_layout(Gtk.ButtonBoxStyle.END)
        save = Button(_("_Save"), Icons.DOCUMENT_SAVE)
        self.save = save
        connect_obj(save, "clicked", self.__save_files, prop, model, library)
        revert = Button(_("_Revert"), Icons.DOCUMENT_REVERT)
        self.revert = revert
        bbox.pack_start(revert, True, True, 0)
        bbox.pack_start(save, True, True, 0)
        self.pack_start(bbox, False, True, 0)

        preview_args = [spin_start, spin_total, model, save, revert]
        preview.connect("clicked", self.__preview_tracks, *preview_args)
        connect_obj(revert, "clicked", self.__update, None, *preview_args[1:])
        spin_total.connect("value-changed", self.__preview_tracks, *preview_args)
        spin_start.connect("value-changed", self.__preview_tracks, *preview_args)
        connect_obj(
            view, "drag-end", self.__class__.__preview_tracks, self, *preview_args
        )
        render.connect("edited", self.__row_edited, model, preview, save)

        connect_obj(
            prop,
            "changed",
            self.__class__.__update,
            self,
            spin_total,
            model,
            save,
            revert,
        )

        for child in self.get_children():
            child.show_all()

    def __row_edited(self, render, path, new, model, preview, save):
        path = Gtk.TreePath.new_from_string(path)
        row = model[path]
        entry = row[0]
        if entry.tracknumber != new:
            entry.tracknumber = new
            preview.set_sensitive(True)
            save.set_sensitive(True)
            model.path_changed(path)

    def __save_files(self, parent, model, library):
        win = WritingWindow(parent, len(model))
        was_changed = set()
        all_done = False
        for entry in model.values():
            song, track = entry.song, entry.tracknumber
            if song.get("tracknumber") == track:
                win.step()
                continue
            if not song.valid():
                win.hide()
                dialog = OverwriteWarning(self, song)
                resp = dialog.run()
                win.show()
                if resp != OverwriteWarning.RESPONSE_SAVE:
                    break
            song["tracknumber"] = track
            try:
                song.write()
            except AudioFileError:
                util.print_exc()
                WriteFailedError(self, song).run()
                library.reload(song, changed=was_changed)
                break
            was_changed.add(song)
            if win.step():
                break
        else:
            all_done = True

        library.changed(was_changed)
        win.destroy()
        self.save.set_sensitive(not all_done)
        self.revert.set_sensitive(not all_done)

    def __preview_tracks(self, ctx, start, total, model, save, revert):
        start = start.get_value_as_int()
        total = total.get_value_as_int()
        for row in model:
            if total:
                s = "%d/%d" % (row.path.get_indices()[0] + start, total)
            else:
                s = str(row.path.get_indices()[0] + start)
            entry = row[0]
            entry.tracknumber = s
            model.row_changed(row.path, row.iter)

        save.set_sensitive(True)
        revert.set_sensitive(True)

    def __update(self, songs, total, model, save, revert):
        if songs is None:
            songs = [e.song for e in model.values()]
        else:
            songs = list(songs)

        def sort_key(song):
            return song("~#track", 0), song("~basename"), song

        songs.sort(key=sort_key)

        model.clear()
        total.set_value(len(songs))

        for song in songs:
            if not song.can_change("tracknumber"):
                self.set_sensitive(False)
                break
        else:
            self.set_sensitive(True)

        for song in songs:
            model.append([Entry(song)])

        save.set_sensitive(False)
        revert.set_sensitive(False)



# Copyright 2016 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

try:
    import musicbrainzngs
except ImportError as e:
    from quodlibet import plugins

    raise plugins.MissingModulePluginError("musicbrainzngs") from e

from quodlibet import app
from quodlibet import const
from quodlibet import util


VARIOUS_ARTISTS_ARTISTID = "89ad4ac3-39f7-470e-963a-56509c546377"


def is_mbid(id_):
    return len(id_) == 36


def search_releases(query):
    """Returns a list of ReleaseResult or raises MusicBrainzError"""

    musicbrainzngs.set_useragent(app.name, const.VERSION)
    return [Release(r) for r in musicbrainzngs.search_releases(query)["release-list"]]


def _get_release(release_id):
    """Returns a release containing all recordings and artists or raises
    MusicBrainzError
    """

    assert is_mbid(release_id)

    return musicbrainzngs.get_release_by_id(
        release_id, includes=["recordings", "artists", "artist-credits", "labels"]
    )["release"]


class Artist:
    def __init__(self, name, sort_name, id_):
        self.name = name
        self.sort_name = sort_name
        self.id = id_  # MusicBrainz Artist ID / Album Artist ID

    @property
    def is_various(self):
        return self.id == VARIOUS_ARTISTS_ARTISTID

    @classmethod
    def from_credit(cls, mbcredit):
        artists = []
        for credit in mbcredit:
            try:
                artist = credit["artist"]
            except TypeError:
                # join strings
                pass
            else:
                artists.append(
                    Artist(artist["name"], artist["sort-name"], artist["id"])
                )
        return artists


class ReleaseTrack:
    """Part of a Release, combines per track and per medium data"""

    def __init__(self, mbtrack, discnumber, track_count, disctitle):
        self._mbtrack = mbtrack
        self.discnumber = discnumber
        self.track_count = track_count
        self.disctitle = disctitle

    @property
    def id(self):
        """MusicBrainz release track ID"""

        return self._mbtrack["id"]

    @property
    def artists(self):
        return Artist.from_credit(self._mbtrack["artist-credit"])

    @property
    def title(self):
        return self._mbtrack["recording"]["title"]

    @property
    def tracknumber(self):
        return self._mbtrack["position"]


class Release:
    def __init__(self, mbrelease):
        self._mbrelease = mbrelease

    @property
    def labelid(self):
        label_list = self._mbrelease.get("label-info-list", [])
        if not label_list:
            return ""
        return label_list[0].get("catalog-number", "")

    @property
    def id(self):
        """MusicBrainz release ID"""

        return self._mbrelease["id"]

    @property
    def date(self):
        return self._mbrelease.get("date", "")

    @property
    def medium_format(self):
        formats = []
        for medium in self._mbrelease["medium-list"]:
            format_ = medium.get("format", "")
            if format_:
                formats.append(format_)
        formats = util.list_unique(formats)
        return "/".join(formats)

    @property
    def country(self):
        return self._mbrelease.get("country", "")

    @property
    def disc_count(self):
        return self._mbrelease["medium-count"]

    @property
    def track_count(self):
        """Number of tracks for all included mediums"""

        track_count = 0
        for medium in self._mbrelease["medium-list"]:
            if "track-count" in medium:
                track_count += medium["track-count"]
            if "pregap" in medium:
                track_count += 1
        return track_count

    @property
    def tracks(self):
        tracks = []
        for medium in self._mbrelease["medium-list"]:
            disc = medium["position"]
            title = medium.get("title", "")
            track_count = medium["track-count"]
            if "pregap" in medium:
                track_count += 1
                tracks.append(ReleaseTrack(medium["pregap"], disc, track_count, title))
            for track in medium["track-list"]:
                tracks.append(ReleaseTrack(track, disc, track_count, title))
        return tracks

    @property
    def title(self):
        return self._mbrelease["title"]

    @property
    def is_single_artist(self):
        """If all tracks have the same artists as the release"""

        ids = [a.id for a in self.artists]
        for track in self.tracks:
            track_ids = [a.id for a in track.artists]
            if ids != track_ids:
                return False
        return True

    @property
    def is_various_artists(self):
        artists = self.artists
        return len(artists) == 1 and artists[0].is_various

    @property
    def artists(self):
        return Artist.from_credit(self._mbrelease["artist-credit"])

    def fetch_full(self):
        """Returns a new Release instance containing more info
        or raises MusicBrainzError

        This method is blocking...
        """

        return Release(_get_release(self.id))



# Copyright 2021 Joschua Gandert
#           2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import datetime
import json
import os.path
import uuid
from collections.abc import Callable, MutableMapping
from pathlib import Path
from typing import Any, NamedTuple

from gi.repository import Gtk
from senf import path2fsn

import quodlibet
from quodlibet import _, app, print_d, print_e, qltk, util
from quodlibet.formats._audio import MIGRATE, AudioFile
from quodlibet.plugins import BoolConfProp, FloatConfProp, PluginConfig
from quodlibet.plugins.songshelpers import each_song, is_finite, is_writable
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.qltk import Icons, SeparatorMenuItem
from quodlibet.qltk.matchdialog import ColumnSpec, MatchListsDialog
from quodlibet.qltk.msg import ErrorMessage, WarningMessage
from quodlibet.qltk.showfiles import show_files
from quodlibet.util import connect_obj, print_exc
from quodlibet.util.matcher import ObjectListMatcher
from quodlibet.util.path import (
    extension_of_file_name,
    join_path_with_escaped_name_of_legal_length,
    stem_of_file_name,
)
from quodlibet.util.songwrapper import SongWrapper, check_wrapper_changed

__all__ = ["ImportExportTagsAndTrackUserDataPlugin"]

_PLUGIN_ID = "ImportExportTagsAndTrackUserData"

# We use this instead of ~playlists, since we want to store playlists in a list
PLAYLISTS_KEY = "//playlists"
IDENTIFIER_KEY = "//identifier"
FILE_STEM_KEY = "//file_stem"

USER_DATA_KEYS = " ".join(MIGRATE | {PLAYLISTS_KEY})

# Could be made configurable in the future, but these likely suffice
# "tag1 tag2" will result in only tag1 and tag2 being exported
# * means all tags (so tags like 'title' and internal (user data) "tags" like ~#added)
# * tag1 tag2 means all except tag1 and tag2
EXPORT_OPTIONS = [
    (_("Export User Data"), USER_DATA_KEYS),
    (_("Export Tags"), f"* {USER_DATA_KEYS} {FILE_STEM_KEY}"),
    (_("Export Tags and User Data"), f"* {FILE_STEM_KEY}"),
    (_("Export File Stems and User Data"), f"{USER_DATA_KEYS} {FILE_STEM_KEY}"),
    (_("Export File Stems and Tags"), f"* {USER_DATA_KEYS}"),
    (_("Export File Stems, Tags and User Data"), "*"),
]

EXPORT_DIR_PATH = Path(quodlibet.get_cache_dir(), "tags_and_track_user_data")
os.makedirs(EXPORT_DIR_PATH, exist_ok=True)

EXPORT_EXTENSION = "json"
TAGS_AND_USERDATA_INDEX_FILE_PATH = EXPORT_DIR_PATH / f"index.{EXPORT_EXTENSION}"


def move_export_to_used(export_path: Path):
    used_path = EXPORT_DIR_PATH / "used"
    used_path.mkdir(exist_ok=True)

    export_path.rename(used_path / export_path.name)


class AlbumId(NamedTuple):
    id_value: str
    title: str
    artist: str
    discs: int
    tracks: int
    last_directory_parts: str

    @classmethod
    def of_song(cls, s: SongWrapper):
        # We're using the last two parts, since sometimes albums have disc folders
        # below the folder that's named after the album
        parts = s("~dirname").rsplit(os.path.sep, maxsplit=2)[-2:]

        return AlbumId(
            s.album_key[0],
            s("albumsort", "") or s("album"),
            s("albumartistsort", "") or s("albumartist") or s("artist"),
            s("~#discs", 1),
            s("~#tracks", 1),
            os.path.join(*parts),
        )


class TrackId(NamedTuple):
    """
    A TrackId is class that's used to identify and match tracks. In theory, almost
    all metadata could be useful for that, but some things are far more relevant than
    others (title vs bpm), and for performance reasons we have to limit it anyway.
    """

    artist: str
    title: str
    disc: int
    discs: int
    track: int
    tracks: int
    file_name: str

    @classmethod
    def of_song(cls, s: SongWrapper | AudioFile):
        return TrackId(
            s("artist"),
            s("title"),
            s("~#disc", 1),
            s("~#discs", 1),
            s("~#track", 1),
            s("~#tracks", 1),
            s("~basename"),
        )

    @property
    def file_stem(self):
        return os.path.splitext(self.file_name)[0]

    @property
    def track_text(self):
        if self.tracks <= 1:
            return str(self.track)
        return f"{self.track}/{self.tracks}"

    @property
    def disc_text(self):
        if self.discs <= 1:
            return str(self.disc)
        return f"{self.disc}/{self.discs}"


class Config:
    _config = PluginConfig(_PLUGIN_ID)

    need_user_check_if_number_of_albums_differs = BoolConfProp(
        _config, "need_user_check_if_number_of_albums_differs", True
    )

    need_user_check_if_number_of_tracks_differs = BoolConfProp(
        _config, "need_user_check_if_number_of_tracks_differs", True
    )

    max_track_similarity_to_need_user_check = FloatConfProp(
        _config, "max_track_similarity_to_need_user_check", 0.76
    )

    max_album_similarity_to_need_user_check = FloatConfProp(
        _config, "max_album_similarity_to_need_user_check", 0.80
    )

    delete_exports_after_importing = BoolConfProp(
        _config, "delete_exports_after_importing", True
    )

    pretty_print_json = BoolConfProp(_config, "pretty_print_json", False)


CONFIG = Config()


class ImportExportTagsAndTrackUserDataPlugin(SongsMenuPlugin):
    PLUGIN_ID = _PLUGIN_ID
    PLUGIN_NAME = _("Import / Export")
    PLUGIN_DESC = _("Imports and exports tags and track user data.")
    PLUGIN_ICON = Icons.EDIT_COPY

    plugin_handles = each_song(is_finite)

    _album_id_to_export_path: MutableMapping[AlbumId, Path]

    def PluginPreferences(self, *args):
        vbox = Gtk.VBox(spacing=6)

        def asd_toggled(button, *args):
            CONFIG.need_user_check_if_number_of_albums_differs = button.get_active()

        def tsd_toggled(button, *args):
            CONFIG.need_user_check_if_number_of_tracks_differs = button.get_active()

        def de_toggled(button, *args):
            CONFIG.delete_exports_after_importing = button.get_active()

        def pp_toggled(button, *args):
            CONFIG.pretty_print_json = button.get_active()

        def mt_scale_changed(scale):
            CONFIG.max_track_similarity_to_need_user_check = scale.get_value()

        def ma_scale_changed(scale):
            CONFIG.max_album_similarity_to_need_user_check = scale.get_value()

        info_box = Gtk.VBox(spacing=6)
        info_frame = qltk.Frame(_("Further information"), child=info_box)
        vbox.pack_start(info_frame, False, True, 0)

        meta_markup = util.monospace(", ".join(MIGRATE))
        info_text = (
            _(
                "The term 'track user data' includes the playlists in which the "
                "selected tracks are and the following metadata:\n\n%s\n"
                "\nBe aware that whatever you chose to export will be imported. "
                "If you exported the file stems (file names without extension), "
                "then, on import, the selected files will be renamed.\n\nAfter "
                "exporting an album you can import the data into another version "
                "of the album. Order and number of tracks can be different. "
                "The plugin matches the exported data to the new tracks, even if "
                "the names of the tracks are slightly different. The automatic "
                "matching is not always correct, so it is better to not reduce "
                "the following similarity values too much."
            )
            % meta_markup
        )

        info_lbl = Gtk.Label(label=info_text, use_markup=True, wrap=True)
        info_box.pack_start(info_lbl, True, True, 0)

        manual_box = Gtk.VBox(spacing=6)
        manual_frame = qltk.Frame(_("User interaction on import"), child=manual_box)
        vbox.pack_start(manual_frame, False, True, 0)

        tsd = Gtk.CheckButton(
            label=_("Require confirmation if number of tracks differs")
        )
        tsd.set_active(CONFIG.need_user_check_if_number_of_tracks_differs)
        tsd.connect("toggled", tsd_toggled)
        manual_box.pack_start(tsd, True, True, 0)

        asd = Gtk.CheckButton(
            label=_("Require confirmation if number of albums differs")
        )
        asd.set_active(CONFIG.need_user_check_if_number_of_albums_differs)
        asd.connect("toggled", asd_toggled)
        manual_box.pack_start(asd, True, True, 0)

        desc = _(
            "Percentage below which the user will have to manually check and "
            "optionally change which track is matched with which."
        )

        perc_table = Gtk.Table(n_rows=2, n_columns=2)
        perc_table.set_col_spacings(6)
        perc_table.set_row_spacings(6)
        manual_box.pack_start(perc_table, True, True, 0)

        def format_perc(scale, value):
            return _("%d %%") % (value * 100)

        def add_perc_scale_with_label(ratio, col, lbl_text, tooltip_text, on_change):
            scale = Gtk.HScale(adjustment=Gtk.Adjustment.new(0, 0, 1, 0.01, 0.01, 0))
            scale.set_digits(2)
            scale.set_tooltip_text(tooltip_text)
            scale.set_value_pos(Gtk.PositionType.RIGHT)
            scale.set_value(ratio)
            scale.connect("format-value", format_perc)
            scale.connect("value-changed", on_change)

            label = Gtk.Label(label=lbl_text)
            label.set_alignment(0.0, 0.5)
            label.set_padding(0, 6)
            label.set_mnemonic_widget(scale)

            xoptions = Gtk.AttachOptions.FILL | Gtk.AttachOptions.SHRINK
            perc_table.attach(label, 0, 1, col, col + 1, xoptions=xoptions)
            perc_table.attach(scale, 1, 2, col, col + 1)

        add_perc_scale_with_label(
            CONFIG.max_track_similarity_to_need_user_check,
            0,
            _("Track similarity:"),
            desc,
            mt_scale_changed,
        )

        add_perc_scale_with_label(
            CONFIG.max_album_similarity_to_need_user_check,
            1,
            _("Album similarity:"),
            desc,
            ma_scale_changed,
        )

        export_box = Gtk.VBox(spacing=6)
        export_frame = qltk.Frame(_("Export files"), child=export_box)
        vbox.pack_start(export_frame, False, True, 0)

        pp = Gtk.CheckButton(label=_("Write pretty and clear JSON (slower)"))
        pp.set_active(CONFIG.pretty_print_json)
        pp.connect("toggled", pp_toggled)
        export_box.pack_start(pp, True, True, 0)

        de = Gtk.CheckButton(label=_("Delete export files after they've been imported"))
        de.set_active(CONFIG.delete_exports_after_importing)
        de.connect("toggled", de_toggled)
        export_box.pack_start(de, True, True, 0)

        return vbox

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._export_collectors = []
        self._import_or_export_option_index = None

        self._album_id_matcher: ObjectListMatcher[AlbumId] = ObjectListMatcher(
            {
                # title is the most reliable
                lambda a: a.title: 9,
                lambda a: a.artist: 4.5,
                lambda a: a.tracks: 1.2,
                # needed in case the album has no tags
                lambda a: a.last_directory_parts: 1,
                # needed in case the album has no tags
                lambda a: a.discs: 0.8,
                # is likely to change unless exact same album
                lambda a: a.id_value: 0.5,
            }
        )
        # We want check similarity afterwards, so it needs be as accurate as possible
        self._album_id_matcher.should_store_similarity_matrix = True
        self._album_id_matcher.should_go_through_every_attribute = True

        self._track_id_matcher: ObjectListMatcher[TrackId] = ObjectListMatcher(
            {
                lambda t: t.title: 8,
                lambda t: t.artist: 3.5,
                lambda t: t.track: 1.2,
                # needed in case the track has no tags
                lambda t: t.file_stem: 1,
                lambda t: t.disc: 0.8,
            }
        )
        self._track_id_matcher.should_store_similarity_matrix = True
        self._album_id_matcher.should_go_through_every_attribute = True

        self._album_id_to_export_path = {}

        submenu = Gtk.Menu()
        self._init_collectors_and_menu(submenu)

        if submenu.get_children():
            self.set_submenu(submenu)
        else:
            self.set_sensitive(False)

    def _init_collectors_and_menu(self, submenu):
        import_item = Gtk.MenuItem(label=_("Import"))
        connect_obj(import_item, "activate", self.__set_import_export_option_index, -1)

        submenu.append(import_item)
        submenu.append(SeparatorMenuItem())

        for idx, (name, query) in enumerate(EXPORT_OPTIONS):
            collector = track_data_collector_for(query)
            self._export_collectors.append(collector)

            item = Gtk.MenuItem(label=name)
            connect_obj(item, "activate", self.__set_import_export_option_index, idx)
            submenu.append(item)

        submenu.append(SeparatorMenuItem())
        open_dir_item = Gtk.MenuItem(label=_("Open Export Directory"))

        def open_export_dir(_):
            show_files(
                path2fsn(EXPORT_DIR_PATH),
                [path2fsn(TAGS_AND_USERDATA_INDEX_FILE_PATH.name)],
            )

        connect_obj(open_dir_item, "activate", open_export_dir, None)
        submenu.append(open_dir_item)

    def __set_import_export_option_index(self, index):
        self._import_or_export_option_index = index

    def _error_msg(self, message):
        title = _("Error in %s") % self.PLUGIN_NAME
        ErrorMessage(app.window, title, message).run()

    def plugin_albums(self, albums):
        index = self._import_or_export_option_index

        if index is None or index >= len(self._export_collectors):
            return

        if index < 0:
            self.import_data_to_albums(albums)
        else:
            collect_data = self._export_collectors[index]
            self.export_albums(albums, collect_data)

        self._rewrite_index()
        self._import_or_export_option_index = None

    def import_data_to_albums(self, albums):
        if not self._try_load_exports():
            return

        for exp_album_id, songs in self._iter_export_album_id_matched_to_songs(albums):
            if exp_album_id is not None:
                self.import_data(exp_album_id, songs)

    def _iter_export_album_id_matched_to_songs(self, albums):
        album_ids = [AlbumId.of_song(songs[0]) for songs in albums]
        exp_album_ids = list(self._album_id_to_export_path.keys())

        exp_indices = self._album_id_matcher.get_indices(album_ids, exp_album_ids)
        size_differs = len(exp_album_ids) != len(exp_album_ids)
        need_check = CONFIG.need_user_check_if_number_of_albums_differs and size_differs

        need_check = need_check or self._does_match_need_manual_check(
            self._album_id_matcher,
            exp_indices,
            CONFIG.max_album_similarity_to_need_user_check,
        )

        if need_check:
            columns = [  #
                ColumnSpec(_("Discs"), lambda a: str(a.discs), False),
                ColumnSpec(_("Tracks"), lambda a: str(a.tracks), False),
                ColumnSpec(_("Title"), lambda a: a.title, True),
                ColumnSpec(_("Artist(s)"), lambda a: a.artist, True),
                ColumnSpec(_("End of path"), lambda a: a.last_directory_parts, True),  #
            ]
            prompt = MatchListsDialog(
                album_ids,
                exp_album_ids,
                exp_indices,
                columns,
                _("Match Albums"),
                _("Continue"),
                id_for_window_tracking=self.PLUGIN_ID,
            )
            exp_indices = prompt.run()

        for exp_idx, songs in zip(exp_indices, albums, strict=False):
            if exp_idx is not None:
                yield exp_album_ids[exp_idx], songs

    def _try_load_exports(self) -> bool:
        """:return: Whether we could load the exports"""

        index_path = TAGS_AND_USERDATA_INDEX_FILE_PATH

        if not index_path.exists():
            self._warning_nothing_to_import()
            return False

        try:
            with index_path.open(encoding="utf-8") as f:
                album_json_key_to_export_file_name = json.load(f)
        except ValueError:
            self._handle_broken_index()
            return False

        if not album_json_key_to_export_file_name:
            self._warning_nothing_to_import()
            return False

        self._load_exports_in_index(album_json_key_to_export_file_name)
        return True

    def _warning_nothing_to_import(self):
        WarningMessage(
            app.window,
            _("Nothing to import"),
            _("You have to export something before you can import."),
        )

    def _load_exports_in_index(self, album_json_key_to_export_file_name):
        for key, file_name in album_json_key_to_export_file_name.items():
            path = EXPORT_DIR_PATH / file_name
            if not path.exists():
                continue

            try:
                # album_id needed to be stored as a json string, since it's a tuple
                album_id = AlbumId(*json.loads(key))
            except ValueError:
                continue

            self._album_id_to_export_path[album_id] = path

    def _handle_broken_index(self):
        index_path = TAGS_AND_USERDATA_INDEX_FILE_PATH

        now = cur_datetime_as_str()
        new_path = index_path.with_name(f"index-broken-{now}.{EXPORT_EXTENSION}")
        index_path.rename(new_path)

        self._error_msg(_("The index was corrupt."))

    def import_data(self, export_album_id: AlbumId, songs: list[SongWrapper]):
        songs = [s for s in songs if is_writable(s)]
        if not songs:
            return
        songs.sort(key=sort_key_for_song)

        export_path = self._album_id_to_export_path[export_album_id]
        changed_songs = self.import_data_and_get_changed(songs, export_path)
        if changed_songs:
            check_wrapper_changed(app.library, changed_songs)

            # Remove used up export
            del self._album_id_to_export_path[export_album_id]
            if CONFIG.delete_exports_after_importing:
                try:
                    export_path.unlink()
                except FileNotFoundError:
                    pass
            else:
                move_export_to_used(export_path)

    def import_data_and_get_changed(
        self, songs: list[SongWrapper], source_path: Path
    ) -> list[SongWrapper]:
        """:return: List of changed songs"""

        exported = self._try_read_source_json(source_path)
        if exported is None:
            return []

        # removes TrackId from exported
        exported_indices = self._get_exported_indices_matched_to_songs(exported, songs)
        if not exported_indices:
            return []

        changed_songs = []
        for song, exp_idx in zip(songs, exported_indices, strict=False):
            if exp_idx is None:
                continue

            self._update_song(exported[exp_idx], song)

            if song._needs_write:
                changed_songs.append(song)

        return changed_songs

    def _try_read_source_json(self, path: Path):
        try:
            print_d(f"Loading from {str(path)!r}")
            with path.open(encoding="utf-8") as f:
                return json.load(f)
        except ValueError:
            print_e(f"Couldn't parse JSON in {path}.")
            self._error_msg(_("Couldn't parse JSON in %s") % path)
            return None
        except OSError:
            print_e(f"Couldn't read {path}")
            self._error_msg(_("Couldn't read %s") % path)
            return None

    def _update_song(self, exported_data, song):
        file_stem = exported_data.pop(FILE_STEM_KEY, None)

        if file_stem is not None:
            file_ext = extension_of_file_name(song("~basename"))

            new_name = f"{file_stem}{file_ext}"
            new_song_path = os.path.join(song("~dirname"), new_name)
            try:
                app.library.rename(song._song, new_song_path)
            except ValueError:
                print_e(f"Could not rename {song._song} to {new_song_path}.")

        for pl_name in exported_data.pop(PLAYLISTS_KEY, []):
            add_song_to_playlist(pl_name, song)

        for tag_key, tag_value in exported_data.items():
            if tag_key in song and song[tag_key] == tag_value:
                continue

            song[tag_key] = tag_value
            song._needs_write = True

    def _rewrite_index(self):
        # AlbumId's are tuples, so we need to serialize them to a string for json
        obj = {json.dumps(k): p.name for k, p in self._album_id_to_export_path.items()}
        self._rewrite_json(obj, TAGS_AND_USERDATA_INDEX_FILE_PATH)

    def _rewrite_json(self, obj, path):
        try:
            print_d(f"Writing to {str(path)!r}")
            with path.open("w+", encoding="utf-8") as f:
                json.dump(obj, f, indent=self._get_json_indent())
        except (ValueError, OSError):
            self._error_msg(_("Couldn't write '%s'") % path)
            print_e(f"Couldn't write {path} due to:")
            print_exc()

    def _get_exported_indices_matched_to_songs(self, exported, songs):
        songs_ids = [TrackId.of_song(s) for s in songs]
        export_ids = [TrackId(*md.pop(IDENTIFIER_KEY)) for md in exported]

        export_ids_indices = self._track_id_matcher.get_indices(songs_ids, export_ids)

        size_differs = len(exported) != len(songs)
        need_check = CONFIG.need_user_check_if_number_of_tracks_differs and size_differs

        need_check = need_check or self._does_match_need_manual_check(
            self._track_id_matcher,
            export_ids_indices,
            CONFIG.max_track_similarity_to_need_user_check,
        )

        if need_check:
            columns = [  #
                ColumnSpec(_("Disc"), lambda t: t.disc_text, False),
                ColumnSpec(_("Track"), lambda t: t.track_text, False),
                ColumnSpec(_("Title"), lambda t: t.title, True),
                ColumnSpec(_("Artist(s)"), lambda t: t.artist, True),
                ColumnSpec(_("File name"), lambda t: t.file_name, True),  #
            ]
            prompt = MatchListsDialog(
                songs_ids,
                export_ids,
                export_ids_indices,
                columns,
                _("Match Tracks"),
                _("Import"),
                id_for_window_tracking=self.PLUGIN_ID,
            )
            return prompt.run()

        return export_ids_indices

    def _does_match_need_manual_check(
        self, matcher, b_indices, max_similarity_to_need_manual_check
    ):
        if max_similarity_to_need_manual_check <= 0.0:
            return False
        if max_similarity_to_need_manual_check >= 1.0:
            return True

        sim_matrix = matcher.similarity_matrix
        for a_idx, b_idx in enumerate(b_indices):
            if b_idx is None:
                continue

            sim = sim_matrix[a_idx][b_idx]
            if sim <= max_similarity_to_need_manual_check:
                return True
        return False

    def _get_json_indent(self):
        return 4 if CONFIG.pretty_print_json else None

    def export_albums(self, albums, collect_data):
        self._try_load_exports()

        for songs in albums:
            self.extract_data_and_export(songs, collect_data)

    def extract_data_and_export(self, songs, collect_data):
        songs.sort(key=sort_key_for_song)
        songs_data = [collect_data(s._song) for s in songs]

        album_id = AlbumId.of_song(songs[0])

        prev_path = self._album_id_to_export_path.get(album_id)

        # this overrides export data with the same album key by design, so a user
        # can simply rerun the export on an album they've modified
        path = new_export_path_for_album(album_id) if prev_path is None else prev_path
        self._album_id_to_export_path[album_id] = path

        self._rewrite_json(songs_data, path)


def cur_datetime_as_str():
    return f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}'


def sort_key_for_song(s: SongWrapper):
    return s("~#disc", 0), s("~#track", 0), s("~basename"), s


def add_song_to_playlist(pl_name, song):
    pl_lib = app.library.playlists
    try:
        pl_lib[pl_name].append(song)
    except KeyError:
        try:
            pl = pl_lib.create(pl_name)
            pl.append(song)
        except ValueError:
            print_e(f"tried to add {song} to playlist {pl_name} but could not due to:")
            print_exc()


TrackData = MutableMapping[str, Any]
"""The term data here includes tags and track user data."""

TrackDataCollector = Callable[[AudioFile], TrackData]


def track_data_collector_for(query: str) -> TrackDataCollector:
    """Creates a callable that returns the track data selected by the query."""
    keys = query.strip().split()

    if keys[0] == "*":
        func = excluding_track_data_collector(set(keys[1:]))
    else:
        func = including_track_data_collector(keys)

    return func


def excluding_track_data_collector(excluded_keys: set[str]) -> TrackDataCollector:
    include_playlist = PLAYLISTS_KEY not in excluded_keys
    include_file_stem = FILE_STEM_KEY not in excluded_keys

    def func(track: AudioFile) -> TrackData:
        md = basic_track_data(track, include_playlist, include_file_stem)
        for key in track:
            if key not in excluded_keys and (key[:1] != "~" or key in MIGRATE):
                md[key] = track[key]

        return md

    return func


def including_track_data_collector(keys: list[str]) -> TrackDataCollector:
    include_playlist = was_removed(keys, PLAYLISTS_KEY)
    include_file_stem = was_removed(keys, FILE_STEM_KEY)

    def func(track: AudioFile) -> TrackData:
        md = basic_track_data(track, include_playlist, include_file_stem)
        for key in keys:
            if key in track:
                md[key] = track[key]
        return md

    return func


def was_removed(elements: list, o: Any) -> bool:
    """:return: whether o was in elements and thus removed from elements"""
    for idx, e in enumerate(elements):
        if e == o:
            del elements[idx]
            return True
    return False


def basic_track_data(
    track: AudioFile, include_playlist: bool, include_file_stem: bool
) -> TrackData:
    md = {IDENTIFIER_KEY: TrackId.of_song(track)}
    if include_playlist:
        if app.library is None:
            raise ValueError("app.library was None - plugin needs it")

        pl_lib = app.library.playlists
        playlist_names = [pl.name for pl in pl_lib.playlists_featuring(track)]
        if playlist_names:
            md[PLAYLISTS_KEY] = playlist_names

    if include_file_stem:
        md[FILE_STEM_KEY] = stem_of_file_name(track("~basename"))
    return md


def new_export_path_for_album(album_id: AlbumId) -> Path:
    stem = f"{album_id.title} - {album_id.artist}"
    path = Path(
        join_path_with_escaped_name_of_legal_length(
            str(EXPORT_DIR_PATH), stem, EXPORT_EXTENSION
        )
    )
    trim_count = 1
    while path.exists():
        new_stem = path.stem[:-trim_count] + uuid.uuid4().hex[:trim_count]
        trim_count += 1
        path = path.with_name(f"{new_stem}.{EXPORT_EXTENSION}")

    return path



# Copyright 2013 Christoph Reiter
#        2020-23 Nick Boultbee
#           2021 Jej@github
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import re

from quodlibet.formats import TIME_TAGS
from quodlibet import config
from quodlibet import util
from quodlibet.formats import AudioFile
from quodlibet.pattern import XMLFromMarkupPattern as XMLFromPattern
from quodlibet.util.string.date import format_date


class PaneConfig:
    """Row pattern format: 'categorize_pattern:display_pattern'

    * display_pattern is optional (fallback: ~#tracks)
    * patterns, tied and normal tags.
    * display patterns can have function prefixes for numerical tags.
    * ':' has to be escaped ('\\:')

    TODO: sort pattern, filter query
    """

    def __init__(self, row_pattern):
        parts = [p.replace(r"\:", ":") for p in (re.split(r"(?<!\\):", row_pattern))]

        def is_numeric(s):
            return s[:2] == "~#" and "~" not in s[2:]

        def is_pattern(s):
            return "<" in s

        def f_round(s):
            return isinstance(s, float) and f"{s:.2f}" or s

        def is_date(s):
            return s in TIME_TAGS

        disp = (
            parts[1]
            if len(parts) >= 2
            else "[i][span alpha='40%']<~#tracks>[/span][/i]"
        )
        cat = parts[0]

        if is_pattern(cat):
            title = util.pattern(cat, esc=True, markup=True)
            try:
                pc = XMLFromPattern(cat)
            except ValueError:
                pc = XMLFromPattern("")
            tags = pc.tags
            format = pc.format_list
            has_markup = True
        else:
            title = util.tag(cat)
            tags = util.tagsplit(cat)
            has_markup = False
            if is_date(cat):

                def format(song: AudioFile) -> list[tuple[str, str]]:
                    fmt = config.gettext("settings", "datecolumn_timestamp_format")
                    date_str = format_date(song(cat), fmt)
                    return [(date_str, date_str)]
            elif is_numeric(cat):

                def format(song: AudioFile) -> list[tuple[str, str]]:
                    v = str(f_round(song(cat)))
                    return [(v, v)]
            else:

                def format(song: AudioFile) -> list[tuple[str, str]]:
                    return song.list_separate(cat)

        if is_pattern(disp):
            try:
                pd = XMLFromPattern(disp)
            except ValueError:
                pd = XMLFromPattern("")
            format_display = pd.format
        else:
            if is_numeric(disp):

                def format_display(coll):
                    return str(f_round(coll(disp)))
            else:

                def format_display(coll):
                    return util.escape(coll.comma(disp))

        self.title = title
        self.tags = set(tags)
        self.format = format
        self.format_display = format_display
        self.has_markup = has_markup

    def __repr__(self):
        return f"<{self.__class__.__name__} title={self.title!r} tags={self.tags!r}>"


def get_headers():
    # QL <= 2.1 saved the headers tab-separated, but had a space-separated
    # default value, so check for that.
    headers = config.get("browsers", "panes")
    if headers == "~people album":
        return headers.split()
    return headers.split("\t")


def save_headers(headers):
    headers = "\t".join(headers)
    config.set("browsers", "panes", headers)



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

import quodlibet
from quodlibet import const


class AboutDialog(Gtk.AboutDialog):
    def __init__(self, parent, app):
        super().__init__()
        self.set_transient_for(parent)
        self.set_program_name(app.name)
        self.set_version(quodlibet.get_build_description())
        self.set_authors(const.AUTHORS)
        self.set_artists(const.ARTISTS)
        self.set_logo_icon_name(app.icon_name)
        self.set_comments(app.description)
        self.set_license_type(Gtk.License.GPL_2_0)
        self.set_translator_credits("\n".join(const.TRANSLATORS))
        self.set_website(const.WEBSITE)
        self.set_copyright(f"{const.COPYRIGHT}\n{const.SUPPORT_EMAIL}")
